import igraph


class IndexTree(igraph.Graph):

    def __init__(self, links, coords):

        super().__init__(directed=True)

        self.add_vertices(coords.shape[0])
        self.add_edges(links)

        self.coords = coords

    # Kom på ett bra namn.
    def walk(self, distance, direction):

        """ Returns all compartments within distance in direction = None, Proximal, Distal """

        pass

    def get_path(self):

        pass

    def verify(self):
        # Check SWC file is not circular
        pass

    def find_parent(self):

        pass

    def find_children(self):

        pass



    # TODO: Funktionalitet vi vill ha
    #
    # -


class IndexLookupError(Exception):
    """Must set a unique value in a BijectiveMap."""

    def __init__(self, value):
        self.value = value
        msg = 'The value "{}" is already in the mapping.'
        super().__init__(msg.format(value))


#  Credit to : https://stackoverflow.com/a/34460187

class IndexLookup(dict):
    """Invertible map."""

    def __init__(self, inverse=None):
        if inverse is None:
            inverse = self.__class__(inverse=self)
        self.inverse = inverse

    def __setitem__(self, key, value):
        if value in self.inverse:
            raise IndexLookupError(value)

        self.inverse._set_item(value, key)
        self._set_item(key, value)

    def __delitem__(self, key):
        self.inverse._del_item(self[key])
        self._del_item(key)

    def _del_item(self, key):
        super().__delitem__(key)

    def _set_item(self, key, value):
        super().__setitem__(key, value)





import ast
import collections
import glob
import json
import os
import sys
import timeit

# Must be run before NEURON import to run in parallel
from mpi4py import MPI

import h5py
import matplotlib.pyplot as plt
import numpy as np
import copy

from snudda.core import Snudda
from snudda.init.init import SnuddaInit
from snudda.input.input import SnuddaInput
from snudda.neurons.neuron_prototype import NeuronPrototype
from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.simulate.simulate import SnuddaSimulate
from snudda.utils import SnuddaLoadNetworkSimulation
from snudda.utils.load import SnuddaLoad
from snudda.utils.snudda_path import snudda_isdir, snudda_parse_path, snudda_simplify_path, get_snudda_data


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


class InputTuning(object):

    def __init__(self, network_path, snudda_data=None, rc=None, input_seed_list=None):

        self.network_path = network_path
        self.neurons_path = None
        self.snudda_data = get_snudda_data(snudda_data=snudda_data)

        self.neuron_types = None
        self.neuron_id = None
        self.input_info = None
        self.neuron_info = None
        self.init_rng = None
        # TODO: Check baseline, close to threshold...
        # TODO: Check when at tonic activity, how sharp short burst can we get without depolarisation block
        self.frequency_range = None
        self.input_duration = None
        self.max_time = None  # self.input_duration * len(self.frequency_range)
        self.num_replicas = None

        self.rc = rc

        if not os.path.isdir(self.network_path):
            os.makedirs(self.network_path)

        self.network_config_file_name = os.path.join(self.network_path, "network-config.json")
        self.network_file = os.path.join(self.network_path, "network-synapses.hdf5")
        self.input_config_file = os.path.join(self.network_path, "input_config.json")

        self.input_seed_list = input_seed_list

        if self.input_seed_list is None:
            self.input_spikes_file = os.path.join(self.network_path, 'input.hdf5')
            self.output_file = os.path.join(self.network_path, "simulations", "output.hdf5")
        else:
            self.input_spikes_file = [os.path.join(self.network_path, f"input-{s}.hdf5") for s in self.input_seed_list]
            self.output_file = [os.path.join(self.network_path, "simulations", f"output-{s}.hdf5") for s in self.input_seed_list]

        self.core = Snudda(self.network_path)
        self.init_helper = SnuddaInit(network_path=self.network_path, snudda_data=self.snudda_data)

    # Writes config files

    def setup_network(self, neurons_path=None, num_replicas=10, neuron_types=None,
                      parameter_key=None, morphology_key=None, modulation_key=None,
                      single_neuron_path=None):

        if not morphology_key and not parameter_key and not modulation_key:
            all_combinations = True
        else:
            all_combinations = False

        self.num_replicas = num_replicas

        # TODO: num_replicas should be set by a parameter, it affects how many duplicates of each neuron
        # and thus how many steps we have between n_min and n_max number of inputs specified.
        config_def = self.create_network_config(neurons_path=neurons_path,
                                                snudda_data=self.snudda_data,
                                                num_replicas=num_replicas,
                                                neuron_types=neuron_types,
                                                single_neuron_path=single_neuron_path,
                                                parameter_key=parameter_key,
                                                morphology_key=morphology_key,
                                                modulation_key=modulation_key,
                                                all_combinations=all_combinations)

        print(f"Writing network config file to {self.network_config_file_name}")
        with open(self.network_config_file_name, "w") as f:
            json.dump(config_def, f, indent=2, cls=NumpyEncoder)

        create_cube_mesh(os.path.join("data", "mesh", "InputTestMesh.obj"), [0, 0, 0], 1e-3,
                         description="Mesh file used for Input Scaling")

        # Write the neurons path to file
        self.write_tuning_info()

        from snudda.place.place import SnuddaPlace
        from snudda.detect.detect import SnuddaDetect
        from snudda.detect.prune import SnuddaPrune

        sp = SnuddaPlace(network_path=self.network_path)
        sp.parse_config(resort_neurons=False)  # By not resorting neurons, we have original order
        sp.write_data()

        sd = SnuddaDetect(network_path=self.network_path, rc=self.rc)
        sd.detect()

        sp = SnuddaPrune(network_path=self.network_path, rc=self.rc)
        sp.prune()

        # TODO: Skip placing neurons that will not receive any inputs or distribute any inputs

    def setup_input(self, input_type=None,
                    num_input_min=100, num_input_max=1000, num_input_steps=None,
                    input_duration=10,
                    input_frequency_range=None,
                    input_correlation=None,
                    use_meta_input=True, generate=True, clear_old_input=True):

        if clear_old_input:
            self.input_info = None

        if not input_frequency_range:
            input_frequency_range = [1.0]

        if not num_input_steps:
            num_input_steps = self.num_replicas

        self.frequency_range = np.array(input_frequency_range)
        self.input_duration = input_duration
        self.max_time = self.input_duration * len(self.frequency_range)

        synapse_density_cortical_input = "1.15*0.05/(1+exp(-(d-30e-6)/5e-6))"
        synapse_density_thalamic_input = "0.05*exp(-d/200e-6)"
        #  synapse_density_thalamic_input = "(d > 100e-6)*1"  # TEST!!

        # TODO: These should be read from JSON file, so user can add additional neuron and input types
        cortical_SPN_synapse_parameter_file = "$DATA/synapses/striatum/M1RH_Analysis_190925.h5-parameters-MS.json"
        cortical_contralateral_SPN_synapse_parameter_file = "$DATA/synapses/striatum/M1LH_Analysis_191001.h5-parameters-MS.json"
        thalamic_SPN_synapse_parameter_file = "$DATA/synapses/striatum/TH_Analysis_191001.h5-parameters-MS.json"

        cortical_FS_synapse_parameter_file = "$DATA/synapses/striatum/M1RH_Analysis_190925.h5-parameters-FS.json"
        cortical_contralateral_FS_synapse_parameter_file = "$DATA/synapses/striatum/M1LH_Analysis_191001.h5-parameters-FS.json"
        thalamic_FS_synapse_parameter_file = "$DATA/synapses/striatum/TH_Analysis_191001.h5-parameters-FS.json"

        cortical_ChIN_synapse_parameter_file = "$DATA/synapses/striatum/M1RH_Analysis_190925.h5-parameters-CHAT.json"
        thalamic_ChIN_synapse_parameter_file = "$DATA/synapses/striatum/TH_Analysis_191001.h5-parameters-CHAT.json"

        cortical_LTS_synapse_parameter_file = "$DATA/synapses/striatum/M1RH_Analysis_190925.h5-parameters-LTS.json"

        if "cortical_contralateral" in input_type.lower():
            synapse_density = synapse_density_cortical_input
            synapse_parameter_file = {"dspn": cortical_contralateral_SPN_synapse_parameter_file,
                                      "ispn": cortical_contralateral_SPN_synapse_parameter_file,
                                      "fs": cortical_contralateral_FS_synapse_parameter_file}

        elif 'cortical' in input_type.lower():
            synapse_density = synapse_density_cortical_input
            synapse_parameter_file = {"dspn": cortical_SPN_synapse_parameter_file,
                                      "ispn": cortical_SPN_synapse_parameter_file,
                                      "fs": cortical_FS_synapse_parameter_file,
                                      "lts": cortical_LTS_synapse_parameter_file,
                                      "chin": cortical_ChIN_synapse_parameter_file}
            print("Using cortical synapse density for input.")
        elif 'thalamic' in input_type.lower():
            synapse_density = synapse_density_thalamic_input
            synapse_parameter_file = {"dspn": thalamic_SPN_synapse_parameter_file,
                                      "ispn": thalamic_SPN_synapse_parameter_file,
                                      "fs": thalamic_FS_synapse_parameter_file,
                                      "chin": thalamic_ChIN_synapse_parameter_file}
            print("Using thalamic synapse density for input")
        else:
            synapse_density = "1"
            synapse_parameter_file = {}
            print("No density profile used for input    .")

        if "cluster" in input_type.lower():
            cluster_size = 4
            cluster_spread = 20e-6
        else:
            cluster_size = None
            cluster_spread = None

        self.create_input_config(input_config_file=self.input_config_file,
                                 input_type=input_type,
                                 input_frequency=list(self.frequency_range),  # [1.0],
                                 input_correlation=input_correlation,
                                 n_input_min=num_input_min,
                                 n_input_max=num_input_max,
                                 num_input_steps=num_input_steps,
                                 synapse_conductance=0.5e-9,
                                 synapse_density=synapse_density,
                                 input_duration=self.input_duration,
                                 synapse_parameter_file=synapse_parameter_file,
                                 cluster_size=cluster_size,
                                 cluster_spread=cluster_spread)

        if generate:
            self.generate_input_helper(use_meta_input=use_meta_input)

        # Info we need to run right duration of simulation
        self.write_tuning_info()

    def generate_input_helper(self, use_meta_input=True):
        if self.input_seed_list is None:
            si = SnuddaInput(input_config_file=self.input_config_file,
                             hdf5_network_file=os.path.join(self.network_path, 'network-synapses.hdf5'),
                             spike_data_filename=self.input_spikes_file,
                             time=self.max_time,
                             logfile=os.path.join(self.network_path, "log", "input.txt"),
                             use_meta_input=use_meta_input, rc=self.rc)
            si.generate()
        else:
            for ctr, (seed, input_file) in enumerate(zip(self.input_seed_list, self.input_spikes_file)):
                print(f"Iteration {ctr}/{len(self.input_seed_list)} (seed: {seed})")

                si = SnuddaInput(input_config_file=self.input_config_file,
                                 hdf5_network_file=os.path.join(self.network_path, 'network-synapses.hdf5'),
                                 spike_data_filename=input_file,
                                 time=self.max_time,
                                 logfile=os.path.join(self.network_path, "log", f"input-{seed}.txt"),
                                 use_meta_input=use_meta_input, rc=self.rc, random_seed=seed)
                si.generate()

    def setup_background_input(self, input_types=["cortical_background", "thalamic_background"],
                               input_density=["1.15*0.05/(1+exp(-(d-30e-6)/5e-6))", "0.05*exp(-d/200e-6)"],
                               input_fraction=[0.5, 0.5],
                               num_input_min=10, num_input_max=500,
                               input_frequency=[1, 1], input_duration=10,
                               generate_input=True):

        """ Tries to find the maximum number of synapses that will not make the neuron spike. """

        if np.sum(input_fraction) != 1:
            # If [0.5, 0.5] and 100 inputs, means 50 and 50 synapses for the two inputs
            input_fraction = np.array(input_fraction) / np.sum(input_fraction)
            print(f"Adjusting input fraction to make total 1: {input_fraction}")

        num_input_steps = self.num_replicas
        self.input_duration = input_duration
        self.max_time = self.input_duration   # Only one frequency for background, we are varying number of inputs

        # TODO: Setup network, run neurons at one frequency.

        # Make sure input is cleared on first iteration, and generated on last iteration
        generate_flag = np.zeros((len(input_frequency),), dtype=bool)
        generate_flag[-1] = generate_input
        clear_flag = np.zeros((len(input_frequency),), dtype=bool)
        clear_flag[0] = True

        for input_type, density, fraction, frequency, gf, cf \
            in zip(input_types, input_density, input_fraction, input_frequency,
                   generate_flag, clear_flag):

            self.setup_input(input_type=input_type,
                             num_input_min=np.round(fraction*num_input_min),
                             num_input_max=np.round(fraction*num_input_max),
                             input_duration=input_duration,
                             input_frequency_range=[frequency],
                             use_meta_input=False,
                             generate=gf,
                             clear_old_input=cf)

    def setup_input_verification(self, input_type="cortical", neuron_type="dSPN",
                                 input_frequency_range=None,
                                 input_duration=10,
                                 generate=True, seed_list=None):

        # Here we use the META input, and replace the frequency, start and end variables
        # input_type can be "cortical", "thalamic", "cortical_background", "thalamic_background"

        if not input_frequency_range:
            input_frequency_range = [1.0, 5.0, 10.0, 15.0, 20.0, 30.0, 40.0]

        self.frequency_range = np.array(input_frequency_range)
        self.input_duration = input_duration
        self.max_time = self.input_duration * len(self.frequency_range)

        n_steps = len(self.frequency_range)
        input_start = input_duration * np.arange(0, n_steps)
        input_end = input_duration * np.arange(1, n_steps + 1)

        input_target = neuron_type
        self.input_info = collections.OrderedDict()
        self.input_info[input_target] = collections.OrderedDict()

        self.input_info[input_target][input_type] = collections.OrderedDict()

        self.input_info[input_target][input_type]["generator"] = "poisson"
        self.input_info[input_target][input_type]["start"] = input_start
        self.input_info[input_target][input_type]["end"] = input_end
        self.input_info[input_target][input_type]["frequency"] = self.frequency_range

        with open(self.input_config_file, "w") as f:
            json.dump(self.input_info, f, indent=4, cls=NumpyEncoder)

        if generate:
            self.generate_input_helper(use_meta_input=True)

        # Info we need to run right duration of simulation
        self.write_tuning_info()

    def analyse_results(self, input_type='', show_plots=False):

        frequency_data, voltage_data = self.load_data()

        self.plot_frequency_data(frequency_data, show_plots=show_plots, input_type_name=input_type)
        self.plot_frequency_data_alt(frequency_data, show_plots=show_plots, input_type_name=input_type)
        self.plot_volt_data(voltage_data, show_plots=show_plots, input_type_name=input_type)
        self.plot_volt_vs_ninputs(voltage_data, show_plots=show_plots, input_type_name=input_type)

        print(f"To plot traces:\n"
              f"python3 plotting/Network_plot_traces.py {self.network_path}output_volt.txt "
              f"{self.network_path}network-synapses.hdf5 ")

    def find_signal_strength(self, requested_frequency=10.0, skip_time=0.0, show_plot=True, quiet_load=False):

        spike_data = dict()
        depolarisation_blocks = dict()

        for idx in range(len(self.input_seed_list)):
            network_info, input_config, _, neuron_id_lookup, neuron_name_list, \
                spike_data[idx], _, time, depolarisation_blocks[idx] = \
                self.load_data_helper(idx=idx, load_input=False, quiet_load=quiet_load)

        input_config_info = dict()

        duration = np.max(time) - skip_time
        requested_spikes = requested_frequency * duration

        bad_neuron_list = []

        for neuron_name in neuron_id_lookup.keys():
            neuron_id = neuron_id_lookup[neuron_name]

            spike_count = np.zeros((len(neuron_id), len(self.input_seed_list)), dtype=int)
            depol_block_flag = np.zeros((len(neuron_id), len(self.input_seed_list)), dtype=bool)

            for idx in range(len(self.input_seed_list)):
                spike_count[:, idx], depol_block_flag[:, idx] = \
                    self.extract_background_spikes(spike_data=spike_data[idx],
                                                   neuron_id=neuron_id,
                                                   skip_time=skip_time,
                                                   depolarisation_blocks=depolarisation_blocks[idx])

            spike_count_mean = np.mean(spike_count, axis=1)

            best_config, neuron_info, depol_block = self.get_best_config(data=spike_count_mean, requested_value=requested_spikes,
                                                                         neuron_id=neuron_id,
                                                                         input_config=input_config,
                                                                         network_info=network_info,
                                                                         depol_block_flag=depol_block_flag)

            if depol_block:
                bad_neuron_list.append(neuron_name)

            input_config_info[neuron_name] = (best_config,
                                              snudda_parse_path(os.path.join(neuron_info["neuron_path"], "meta.json"),
                                                                snudda_data=self.snudda_data),
                                              neuron_info["parameter_key"],
                                              neuron_info["morphology_key"])

            # Just an idiot check to make sure all neurons we are comparing are the same
            for nid in neuron_id:
                assert network_info.data["neurons"][neuron_id[0]]["name"] == network_info.data["neurons"][nid]["name"]

            if depol_block:
                label = f"signal-{requested_frequency}-Hz-BLOCKED"
            else:
                label = f"signal-{requested_frequency}-Hz"

            self.plot_signal_info(neuron_id=neuron_id, neuron_info=neuron_info, best_config=best_config,
                                  spike_count=spike_count, input_config=input_config, max_time=np.max(time),
                                  requested_frequency=requested_frequency, depol_block_flag=depol_block_flag,
                                  label=label, show_plot=show_plot)

        for name in bad_neuron_list:
            print(f"Found early depolarisation block: {name}")

        return input_config_info

    def get_best_config(self, data, requested_value, neuron_id, input_config, network_info,
                        depol_block_flag):

        idx_above = np.argmax(data > requested_value)

        if data[idx_above] < requested_value:
            # If we did not reach the requested value, pick the closest point.
            idx_above = idx_below = np.argmin(abs(data - requested_value))

        elif idx_above == 0:
            idx_below = 0
        else:
            idx_below = idx_above - 1

        config_above = input_config[str(neuron_id[idx_above])]
        config_below = input_config[str(neuron_id[idx_below])]

        input_type = list(config_above.keys())
        assert len(input_type) == 1, f"Interpolation can only handle one input type. Found {input_type}"
        input_type = input_type[0]

        n_syn_above = config_above[input_type]["num_inputs"]
        n_syn_below = config_below[input_type]["num_inputs"]

        value_above = data[idx_above]
        value_below = data[idx_below]

        depol_block = depol_block_flag[:idx_above, :].any()

        assert value_below <= value_above

        if n_syn_above == n_syn_below:
            n_syn = n_syn_above
        else:
            n_syn = int(np.round(n_syn_below
                                 + (n_syn_above - n_syn_below) * (requested_value - value_below)
                                 / (value_above - value_below)))

        try:
            assert n_syn_below <= n_syn <= n_syn_above, f"NOT TRUE: n_syn_below {n_syn_below} <= n_syn {n_syn} <= n_syn_above {n_syn_above}"
        except:
            import traceback
            import pdb
            print(traceback.format_exc())
            pdb.set_trace()

        best_config = copy.deepcopy(config_above)
        best_config[input_type]["num_inputs"] = n_syn

        neuron_info = network_info.data["neurons"][neuron_id[idx_above]]

        return best_config, neuron_info, depol_block

    def find_highest_non_spiking_background_input(self, skip_time=0.0, show_plot=True, quiet_load=False):

        spike_data = dict()
        depolarisation_blocks = dict()

        for idx in range(len(self.input_seed_list)):
            network_info, input_config, _, neuron_id_lookup, neuron_name_list, \
                spike_data[idx], _, time, depolarisation_blocks[idx] = \
                self.load_data_helper(idx=idx, load_input=False, quiet_load=quiet_load)

        input_config_info = dict()

        for neuron_name in neuron_id_lookup.keys():
            neuron_id = neuron_id_lookup[neuron_name]

            spike_count = np.zeros((len(neuron_id), len(self.input_seed_list)), dtype=int)
            depol_block_flag = np.zeros((len(neuron_id), len(self.input_seed_list)), dtype=bool)

            for idx in range(len(self.input_seed_list)):
                spike_count[:, idx], depol_block_flag[:, idx] = \
                    self.extract_background_spikes(spike_data=spike_data[idx],
                                                   neuron_id=neuron_id,
                                                   skip_time=skip_time,
                                                   depolarisation_blocks=depolarisation_blocks[idx])

            spike_count_sum = np.sum(spike_count, axis=1)

            # TODO: Need to handle if there are spikes in all traces
            best_ctr = 0
            for ctr, sc in enumerate(spike_count_sum):
                if sc > 0:
                    break
                else:
                    best_ctr = ctr

            best_neuron_id = neuron_id[best_ctr]

            neuron_info = network_info.data["neurons"][best_neuron_id]
            input_config_info[neuron_name] = (input_config[str(best_neuron_id)],
                                              snudda_parse_path(os.path.join(neuron_info["neuron_path"], "meta.json"),
                                                                snudda_data=self.snudda_data),
                                              neuron_info["parameter_key"],
                                              neuron_info["morphology_key"])

            # Just an idiot check to make sure all neurons we are comparing are the same
            for nid in neuron_id:
                assert network_info.data["neurons"][neuron_id[0]]["name"] == network_info.data["neurons"][nid]["name"]

            self.plot_background_info(neuron_id=neuron_id, neuron_info=neuron_info, best_neuron_id=best_neuron_id,
                                      spike_count=spike_count, input_config=input_config, depol_block_flag=depol_block_flag,
                                      max_time=np.max(time), label="background-inputs", show_plot=show_plot)

        return input_config_info

    def plot_background_info(self, neuron_id, neuron_info, best_neuron_id, spike_count, input_config,
                             max_time, skip_time=0, label="background-inputs", show_plot=True,
                             depol_block_flag=None):

        n_inputs_total = np.zeros((len(neuron_id),), dtype=int)
        fig_dir = os.path.join(self.network_path, "figures")

        if not os.path.isdir(fig_dir):
            os.mkdir(fig_dir)

        fig_name = os.path.join(fig_dir, f"{neuron_info['morphology_key']}-{neuron_info['parameter_key']}-{neuron_info['name']}-{label}.png")

        for ctr, nid in enumerate(neuron_id):
            # Get total input.
            for input_conf in input_config[str(nid)].values():
                n_inputs_total[ctr] += input_conf["num_inputs"]

        best_idx = np.where(neuron_id == best_neuron_id)[0]

        plt.figure()
        plt.plot(n_inputs_total, spike_count/(max_time-skip_time), 'k.')

        if depol_block_flag is not None:
            # Mark the depolarisation blocks with a red circle.
            bad_idx = np.where(depol_block_flag)
            plt.plot(n_inputs_total[bad_idx[0]], spike_count[bad_idx]/(max_time-skip_time), 'ro')

        plt.plot(n_inputs_total[best_idx], spike_count[best_idx]/(max_time-skip_time), 'b*')
        plt.xlabel("Total number of synapses")
        plt.ylabel("Spike frequency")
        plt.title(f"Neuron {neuron_info['name']}")

        print(f"Saving figure to {fig_name}")
        plt.savefig(fig_name)

        if show_plot:
            plt.ion()
            plt.show()

    def plot_signal_info(self, neuron_id, neuron_info, best_config, spike_count, input_config,
                         max_time, requested_frequency, skip_time=0,
                         label="background-inputs", show_plot=True,
                         depol_block_flag=None):

        n_inputs_total = np.zeros((len(neuron_id),), dtype=int)
        fig_dir = os.path.join(self.network_path, "figures")

        if not os.path.isdir(fig_dir):
            os.mkdir(fig_dir)

        fig_name = os.path.join(fig_dir, f"{neuron_info['morphology_key']}-{neuron_info['parameter_key']}-{neuron_info['name']}-{label}.png")

        for ctr, nid in enumerate(neuron_id):
            # Get total input.
            for input_conf in input_config[str(nid)].values():
                n_inputs_total[ctr] += input_conf["num_inputs"]

        plt.figure()
        plt.plot(n_inputs_total, spike_count/(max_time-skip_time), 'k.')

        if depol_block_flag is not None:
            # Mark the depolarisation blocks with a red circle.
            bad_idx = np.where(depol_block_flag)
            plt.plot(n_inputs_total[bad_idx[0]], spike_count[bad_idx]/(max_time-skip_time), 'ro')

        input_type = list(best_config.keys())
        assert len(input_type) == 1
        input_type = input_type[0]
        best_n_syn = best_config[input_type]["num_inputs"]

        y_lim = plt.gca().get_ylim()
        plt.plot([best_n_syn, best_n_syn], y_lim, 'r-')  # best n_inputs

        x_lim = plt.gca().get_xlim()
        plt.plot(x_lim, [requested_frequency, requested_frequency], 'k--')

        # plt.plot(n_inputs_total[best_idx], spike_count[best_idx]/(max_time-skip_time), 'r*')
        plt.xlabel("Total number of synapses")
        plt.ylabel("Spike frequency")
        plt.title(f"Neuron {neuron_info['name']} ({best_n_syn} {input_type} synapses)")

        print(f"Saving figure to {fig_name}")
        plt.savefig(fig_name)

        if show_plot:
            plt.ion()
            plt.show()

    def update_meta(self, input_config_info, overwrite=True, set_frequency=None):

        last_meta_file = None
        meta_data = None

        for input_config, meta_file, parameter_key, morphology_key in input_config_info.values():

            if meta_file != last_meta_file:
                # Write old data to file
                if last_meta_file is not None:
                    print(f"Writing {last_meta_file}")
                    with open(last_meta_file, "w") as f:
                        json.dump(meta_data, f, indent=4)

                # Load new data
                with open(meta_file, "r") as f:
                    meta_data = json.load(f)
                last_meta_file = meta_file

            new_config = copy.deepcopy(input_config)
            try:
                for input_name in new_config.keys():
                    if set_frequency is not None:
                        # Override the frequency (this is useful to have a signal, but not have it active)
                        new_config[input_name]["frequency"] = set_frequency
                    else:
                        # Use the frequency from the input tuning
                        new_config[input_name]["frequency"] = new_config[input_name]["frequency"][0]

                    del new_config[input_name]["start"]
                    del new_config[input_name]["end"]

                    # If parameter_file and parameter-ist both are given, only keep the latter
                    if "parameter_file" in new_config[input_name] and "parameter_list" in new_config[input_name]:
                        del new_config[input_name]["parameter_list"]
            except:
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()

            if parameter_key not in meta_data or morphology_key not in meta_data[parameter_key]:
                print(f"Parameter key {parameter_key}, morphology key {morphology_key} not found in {meta_file} -- was it manually removed?")
                continue

            if overwrite:
                meta_data[parameter_key][morphology_key]["input"] = new_config
            else:
                meta_data[parameter_key][morphology_key]["input"] |= new_config

        # Write last iteration to file
        if last_meta_file is not None:
            print(f"Writing {last_meta_file}")
            with open(last_meta_file, "w") as f:
                json.dump(meta_data, f, indent=4)

    def load_data(self, skip_time=0.0, quiet_load=False):

        input_data = dict()
        spike_data = dict()
        volt = dict()
        time = dict()
        depolarisation_blocks = dict()

        if self.input_seed_list is None:
            network_info, input_config, input_data, neuron_id_lookup, neuron_name_list, \
            spike_data, volt, time, depolarisation_blocks \
                = self.load_data_helper(idx=None, quiet_load=quiet_load)

            first_input_data = input_data["input"]
        else:
            for idx in range(len(self.input_seed_list)):
                network_info, input_config, input_data[idx], neuron_id_lookup, neuron_name_list, \
                    spike_data[idx], volt[idx], time[idx], depolarisation_blocks[idx] \
                    = self.load_data_helper(idx=idx, quiet_load=quiet_load)

            first_input_data = input_data[0]["input"]

        n_inputs_lookup = dict()

        for neuron_label in first_input_data:
            neuron_id = int(neuron_label)
            n_inputs = 0

            for input_type in first_input_data[neuron_label]:
                n_inputs += first_input_data[neuron_label][input_type]["spikes"].shape[0]

            n_inputs_lookup[neuron_id] = n_inputs

        frequency_data = dict()
        voltage_data = dict()

        for neuron_name in neuron_id_lookup.keys():
            frequency_data[neuron_name] = dict()
            voltage_data[neuron_name] = dict()

        for neuron_name in neuron_name_list:
            for neuron_id in neuron_id_lookup[neuron_name]:

                if neuron_id not in n_inputs_lookup:
                    print(f"No inputs for neuron_id={neuron_id}, ignoring. Please update your setup.")
                    continue

                if self.input_seed_list is None:

                    n_inputs = n_inputs_lookup[neuron_id]

                    if n_inputs not in frequency_data[neuron_name]:
                        frequency_data[neuron_name][n_inputs] = dict()

                    frequency_data[neuron_name][n_inputs] = self.extract_frequencies(spike_data=spike_data,
                                                                                     config_data=input_config,
                                                                                     neuron_id=neuron_id,
                                                                                     skip_time=skip_time)
                    voltage_data[neuron_name][n_inputs] = self.extract_voltage(volt=volt,
                                                                               time=time,
                                                                               config_data=input_config,
                                                                               neuron_id=neuron_id,
                                                                               skip_time=skip_time)

                else:

                    for idx in range(len(self.input_seed_list)):

                        n_inputs = n_inputs_lookup[neuron_id]

                        if n_inputs not in frequency_data[neuron_name]:
                            frequency_data[neuron_name][n_inputs] = dict()

                        frequency_data[neuron_name][n_inputs][idx] = self.extract_frequencies(spike_data=spike_data,
                                                                                              config_data=input_config,
                                                                                              neuron_id=neuron_id,
                                                                                              skip_time=skip_time)
                        voltage_data[neuron_name][n_inputs][idx] = self.extract_voltage(volt=volt,
                                                                                        time=time,
                                                                                        config_data=input_config,
                                                                                        neuron_id=neuron_id,
                                                                                        skip_time=skip_time)

        # TODO: Load voltage trace and warn for depolarisation blocking

        return frequency_data, voltage_data

    def load_data_helper(self, idx=None, load_input=True, quiet_load=False):

        network_file = os.path.join(self.network_path, "network-synapses.hdf5")
        network_info = SnuddaLoad(network_file)
        input_data = None

        if idx is None:
            output_file = self.output_file
            if load_input and os.path.isfile(self.input_spikes_file):
                input_data = h5py.File(self.input_spikes_file, "r")
        else:
            output_file = self.output_file[idx]
            if load_input and os.path.isfile(self.input_spikes_file[idx]):
                input_data = h5py.File(self.input_spikes_file[idx], "r")

        output_data_loader = SnuddaLoadNetworkSimulation(network_path=self.network_path,
                                                         network_simulation_output_file=output_file,
                                                         do_test=True, quiet_load=quiet_load)
        spike_data = output_data_loader.get_spikes()

        # cell_id = output_data_loader.get_id_of_neuron_type()
        volt = output_data_loader.get_voltage()
        time = output_data_loader.get_time()
        # We need to figure out what neuronID correspond to that morphologies
        # Then figure out what input frequencies the different runs had

        neuron_id_list = output_data_loader.get_id_of_neuron_type()
        neuron_name_list = output_data_loader.get_neuron_name()

        # For each morphology-model we have a list of the run with that model
        neuron_id_lookup = dict()

        for neuron_id, neuron_name in zip(neuron_id_list, neuron_name_list):
            if neuron_name in neuron_id_lookup:
                neuron_id_lookup[neuron_name].append(neuron_id)
            else:
                neuron_id_lookup[neuron_name] = [neuron_id]

        # Next identify number of inputs each run had
        input_config = self.load_input_config()

        depolarisation_blocks = output_data_loader.get_depolarisation_dictionary()

        return network_info, input_config, input_data, neuron_id_lookup, neuron_name_list, spike_data, volt, time, depolarisation_blocks

    def is_depolarisation_blocked(self, neuron_id, time_range, depolarisation_blocks):

        # Checks if there are any depolarisation blocks for neuron_id within the time_range

        is_blocked = False

        if neuron_id in depolarisation_blocks:

            for start_block, end_block in depolarisation_blocks[neuron_id]:

                if time_range[0] < end_block and time_range[1] > start_block:
                    is_blocked = True

        return is_blocked

    def plot_depolarisation_blocked_neurons(self, freq_bin=10):

        spike_data = dict()
        volt = dict()
        depolarisation_blocks = dict()

        if type(self.input_spikes_file) == list:
            idx_list = np.arange(0, len(self.input_spikes_file))

            for idx in idx_list:
                network_info, input_config, _, neuron_id_lookup, neuron_name_list, \
                    spike_data[idx], volt[idx], time, depolarisation_blocks[idx] = \
                    self.load_data_helper(idx=idx, load_input=False, quiet_load=True)

        else:
            network_info, input_config, _, neuron_id_lookup, neuron_name_list, \
                spike_data[0], volt[0], time, depolarisation_blocks[0] = \
                self.load_data_helper(load_input=False, quiet_load=True)

        bad_neurons = []

        for idx in depolarisation_blocks:
            bad_neurons += list(depolarisation_blocks[idx].keys())

        for neuron_id in bad_neurons:
            plt.figure()

            for ctr, idx in enumerate(volt):
                plt.plot(time*1e3, volt[idx][neuron_id]*1e3)
                plt.plot(spike_data[idx][neuron_id]*1e3,
                         np.full(spike_data[idx][neuron_id].shape, 40+ctr), 'k.')
                if neuron_id in depolarisation_blocks[idx]:
                    for bad_start, bad_end in depolarisation_blocks[idx][neuron_id]:
                        plt.plot([bad_start*1e3, bad_end*1e3], [50, 50], 'r')

            full_morph_key = network_info.data["neurons"][neuron_id]["morphology_key"]
            full_param_key = network_info.data["neurons"][neuron_id]["parameter_key"]
            neuron_type = network_info.data["neurons"][neuron_id]["type"]

            if freq_bin is not None:
                n_bins = int(np.floor(np.max(time) / freq_bin)) + 1
                freq_data = np.zeros((n_bins, len(spike_data.keys())))

                for idx in spike_data:
                    try:
                        unique, counts = np.unique(np.floor(spike_data[idx][neuron_id].flatten() / freq_bin).astype(int), return_counts=True)
                    except:
                        import traceback
                        print(traceback.format_exc())
                        import pdb
                        pdb.set_trace()

                    for val, ctr in zip(unique, counts):
                        freq_data[val, idx] = ctr / freq_bin

                freq_vals = np.mean(freq_data, axis=1)
                for idx, fv in enumerate(freq_vals):
                    plt.text((idx+0.5)*freq_bin*1e3, -10, f"{fv:.1f}Hz",
                             horizontalalignment="center")

            plt.title(f"{neuron_type}, param: {full_param_key}, morph: {full_morph_key} (id {neuron_id})")
            plt.xlabel("Time (ms)")
            plt.ylabel("Voltage (mV)")

            fig_path = os.path.join(self.network_path, "figures",
                                    f"Bad-trace-{neuron_type}-{full_param_key}-{full_morph_key}.png")

            if not os.path.exists(os.path.dirname(fig_path)):
                os.mkdir(os.path.dirname(fig_path))

            plt.savefig(fig_path, dpi=300)

            plt.ion()
            plt.show()

    def plot_voltage_trace(self, morphology_key, parameter_key, mp_idx=None, time_range=None):

        # Find all neurons with the morphology key, and parameter key

        # If idx is given pick the n:th trace specified to plot, if not plot all traces.
        # idx can be a list of multiple traces.

        spike_data = dict()
        volt = dict()
        depolarisation_blocks = dict()

        if type(self.input_spikes_file) == list:
            idx_list = np.arange(0, len(self.input_spikes_file))

            for idx in idx_list:
                network_info, input_config, _, neuron_id_lookup, neuron_name_list, \
                    spike_data[idx], volt[idx], time, depolarisation_blocks[idx] = \
                    self.load_data_helper(idx=idx, load_input=False, quiet_load=True)

        else:
            network_info, input_config, _, neuron_id_lookup, neuron_name_list, \
                spike_data[0], volt[0], time, depolarisation_blocks[0] = \
                self.load_data_helper(load_input=False, quiet_load=True)

        morph_flag = np.array([morphology_key in n["morphology_key"] for n in network_info.data["neurons"]], dtype=bool)
        param_flag = np.array([parameter_key in n["parameter_key"] for n in network_info.data["neurons"]], dtype=bool)

        match_idx = np.where(np.logical_and(morph_flag, param_flag))[0]

        if mp_idx is None:
            plot_idx = match_idx
        else:
            plot_idx = match_idx[mp_idx]

        full_morph_key = network_info.data["neurons"][plot_idx[0]]["morphology_key"]
        full_param_key = network_info.data["neurons"][plot_idx[0]]["parameter_key"]
        neuron_type = network_info.data["neurons"][plot_idx[0]]["type"]

        fig = plt.figure()
        for r_idx in volt.keys():
            for ctr, idx in enumerate(plot_idx):
                spike_times = spike_data[r_idx][idx]

                if time_range is None:
                    plt.plot(time*1e3, volt[r_idx][idx]*1e3)
                    plt.plot(spike_times*1e3, np.full(spike_times.shape, 40+ctr), 'k.')
                else:
                    t_idx = np.logical_and(time_range[0] <= time, time <= time_range[1])
                    plt.plot(time[t_idx]*1e3, volt[r_idx][idx][t_idx]*1e3)
                    s_idx = np.where(np.logical_and(time_range[0] <= spike_times,
                                                    spike_times <= time_range[1]))
                    plt.plot(spike_times[s_idx]*1e3, np.full(spike_times[s_idx].shape, 40+ctr), 'k.')

                if idx in depolarisation_blocks[r_idx]:
                    bad_ranges = depolarisation_blocks[r_idx][idx]
                    for bad_start, bad_end in bad_ranges:
                        if time_range is None or (bad_start <= time_range[1] and bad_end >= time_range[0]):
                            plt.plot([bad_start*1e3, bad_end*1e3], [50, 50], 'r')

        plt.title(f"{neuron_type}, param: {full_param_key}, morpg: {full_morph_key}")
        plt.xlabel("Time (ms)")
        plt.ylabel("Voltage (mV)")

        fig_path = os.path.join(self.network_path, "figures", f"Trace-{neuron_type}-{full_param_key}-{full_morph_key}.png")
        plt.savefig(fig_path)

        plt.ion()
        plt.show()

    # TODO: We should set skip_time to 1 second, 0 for now while testing
    def extract_frequencies(self, spike_data, config_data, neuron_id, skip_time=0.0):

        # This function checks the config_data to find out what time ranges to extract

        assert skip_time >= 0, "Time skipped at beginning of stimulus must be positive or zero"
        assert len(config_data[str(neuron_id)].values()) == 1, "Analysis can only handle one input type at a time"
        input_type = list(config_data[str(neuron_id)].keys())[0]

        cfg_data = config_data[str(neuron_id)][input_type]

        input_frequency = []
        output_frequency = []

        for start_time, end_time, input_freq in zip(cfg_data["start"], cfg_data["end"], cfg_data["frequency"]):

            assert start_time + skip_time < end_time, \
                f"Too large skip time, no data to analyse. start {start_time} + skip {skip_time} < end {end_time}"

            try:
                spike_idx = np.where((start_time + skip_time <= spike_data[neuron_id])
                                     & (spike_data[neuron_id] <= end_time))[0]
                output_freq = len(spike_idx) / (end_time - start_time)
            except:
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()

            input_frequency.append(input_freq)
            output_frequency.append(output_freq)

        input_frequency = np.array(input_frequency)
        output_frequency = np.array(output_frequency)

        return input_frequency, output_frequency, input_type

    def extract_background_spikes(self, spike_data, neuron_id, skip_time=0.0, depolarisation_blocks=None):

        spike_count = np.zeros((len(neuron_id), ), dtype=int)

        for ctr, n_id in enumerate(neuron_id):
            spike_count[ctr] = len(np.where(spike_data[n_id] >= skip_time)[0])

        if depolarisation_blocks is not None:
            depol_block_flag = np.zeros((len(neuron_id),), dtype=bool)
            for ctr, n_id in enumerate(neuron_id):
                if n_id in depolarisation_blocks:
                    depol_block_flag[ctr] = True

            return spike_count, depol_block_flag

        return spike_count

    def load_input_config(self):

        input_config_file = os.path.join(self.network_path, "input_config.json")
        with open(input_config_file) as f:
            input_config = json.load(f, object_pairs_hook=collections.OrderedDict)

        return input_config

    def extract_voltage(self, volt, time, config_data, neuron_id, skip_time=0.0):

        assert skip_time >= 0
        assert len(config_data[str(neuron_id)].values()) == 1

        input_type = list(config_data[str(neuron_id)].keys())[0]
        cfg_data = config_data[str(neuron_id)][input_type]

        input_frequency = []
        mean_voltage = []
        max_voltage = []

        for start_time, end_time, input_freq in zip(cfg_data["start"], cfg_data["end"], cfg_data["frequency"]):
            assert start_time + skip_time < end_time, "Too large skip time, no data to analyse"

            idx = np.where((start_time + skip_time <= time) & (time <= end_time))[0]
            v = volt[neuron_id][idx]

            input_frequency.append(input_freq)
            mean_voltage.append(np.mean(v))
            max_voltage.append(np.max(v))

        input_frequency = np.array(input_frequency)
        mean_voltage = np.array(mean_voltage)
        max_voltage = np.array(max_voltage)

        return input_frequency, mean_voltage, max_voltage, input_type

    def load_spike_data(self, file_name, n_cells):

        data = np.genfromtxt(file_name, delimiter='\t')
        spike_time = data[:, 0] * 1e-3
        spike_id = data[:, 1].astype(int)

        spike_times = dict()
        for nid in range(0, n_cells):
            spike_times[nid] = []

        for sid, t in zip(spike_id, spike_time):
            spike_times[sid].append(t)

        for nid in range(0, n_cells):
            spike_times[nid] = np.array(sorted(spike_times[nid]))

        return spike_times

    def plot_volt_data(self, volt_data, show_plots=True, input_type_name=''):

        for neuron_name in volt_data:
            fig, ax = plt.subplots()
            legend_text = []
            input_type_all = None

            cmap = plt.get_cmap('tab20', len(volt_data[neuron_name]))
            ax.set_prop_cycle('color', [cmap(i) for i in range(0, len(volt_data[neuron_name]))])

            for num_input in volt_data[neuron_name]:
                input_freq, mean_voltage, max_voltage, input_type = volt_data[neuron_name][num_input]

                if input_type_all:
                    assert input_type == input_type_all, "All input types must be the same for neuron"
                else:
                    input_type_all = input_type

                legend_text.append(f"n={num_input}")
                ax.plot(input_freq, mean_voltage, linestyle='-', marker='o')

            plt.title(f"{neuron_name} receiving {input_type_all} input")
            plt.xlabel("Input frequency (per synapse)")
            plt.ylabel("Mean membrane potential (V)")
            ax.legend(legend_text)

            if show_plots:
                plt.ion()
                plt.show()
                plt.pause(0.001)

            # fig_name = os.path.join(self.network_path, "figures", f"input-scaling-freq-{neuron_name}-mean-voltage-{input_type_name}.pdf")
            fig_name = os.path.join(self.network_path, "figures",
                                    f"input-scaling-freq-{neuron_name}-mean-voltage-{input_type_name}.png")
            if not os.path.exists(os.path.dirname(fig_name)):
                os.mkdir(os.path.dirname(fig_name))

            plt.savefig(fig_name, dpi=300)

            if not show_plots:
                plt.close()

    def plot_volt_vs_ninputs(self, volt_data, show_plots=True, input_type_name=''):

        for neuron_name in volt_data:
            fig, ax = plt.subplots()
            legend_text = []
            input_type_all = None

            cmap = plt.get_cmap('tab20', len(volt_data[neuron_name]))
            ax.set_prop_cycle('color', [cmap(i) for i in range(0, len(volt_data[neuron_name]))])
            num_inputs = []
            mean_voltages = []
            for num_input in volt_data[neuron_name]:
                input_freq, mean_voltage, max_voltage, input_type = volt_data[neuron_name][num_input]
                num_inputs.append(num_input)
                mean_voltages.append(mean_voltage)
                if input_type_all:
                    assert input_type == input_type_all, "All input types must be the same for neuron"
                else:
                    input_type_all = input_type

            # legend_text.append(f"input freq.={input_freq[0]}Hz")#Currently assumes one frequency used. Fix later if necessary
            # ax.legend(legend_text)
            ax.plot(num_inputs, mean_voltages, linestyle='None', marker='o')
            # plt.title(f"{neuron_name} receiving {input_type_all} input")
            plt.xlabel("Number of synapses")
            plt.ylabel("Mean membrane potential (V)")

            if show_plots:
                plt.ion()
                plt.show()
                plt.pause(0.001)

            # fig_name = os.path.join(self.network_path, "figures", f"input-scaling-ninputs-{neuron_name}-mean-voltage-{input_type_name}.pdf")
            fig_name = os.path.join(self.network_path, "figures",
                                    f"input-scaling-ninputs-{neuron_name}-mean-voltage-{input_type_name}.png")
            if not os.path.exists(os.path.dirname(fig_name)):
                os.mkdir(os.path.dirname(fig_name))

            plt.savefig(fig_name, bbox_inches='tight', dpi=300)

            if not show_plots:
                plt.close()

    # TODO: Extract spiking frequency (skip first second for each interval to let network settle)
    # TODO: Create summary graphs

    def plot_frequency_data(self, frequency_data, show_plots=True, input_type_name=''):

        for neuron_name in frequency_data:
            fig, ax = plt.subplots()
            legend_text = []
            input_type_all = None

            cmap = plt.get_cmap('tab20', len(frequency_data[neuron_name]))
            ax.set_prop_cycle('color', [cmap(i) for i in range(0, len(frequency_data[neuron_name]))])

            for num_input in frequency_data[neuron_name]:
                input_freq, output_freq, input_type = frequency_data[neuron_name][num_input]

                if input_type_all:
                    assert input_type == input_type_all, "All input types must be the same for neuron"
                else:
                    input_type_all = input_type

                legend_text.append(f"n={num_input}")
                ax.plot(input_freq, output_freq)

            plt.title(f"{neuron_name} receiving {input_type_all} input")
            plt.xlabel("Input frequency (per synapse)")
            plt.ylabel("Firing frequency")
            ax.legend(legend_text)

            if show_plots:
                plt.ion()
                plt.show()
                plt.pause(0.001)

            fig_name = os.path.join(self.network_path, "figures",
                                    f"input-scaling-freq-{neuron_name}-{input_type_name}.png")
            if not os.path.exists(os.path.dirname(fig_name)):
                os.mkdir(os.path.dirname(fig_name))

            plt.savefig(fig_name, dpi=300)

            if not show_plots:
                plt.close()

    def plot_frequency_data_alt(self, frequency_data, show_plots=True, input_type_name=''):

        _freq_data = dict()
        _all_num_inputs = []
        _all_input_freq = []

        input_type_all = None

        # Put everything in a dictionary structure first
        for neuron_name in frequency_data:
            _freq_data[neuron_name] = dict()

            for num_inputs in frequency_data[neuron_name]:
                input_freq, output_freq, input_type = frequency_data[neuron_name][num_inputs]

                if input_type_all is not None:
                    assert input_type == input_type_all, "All neurons must have same input type for this plot"
                else:
                    input_type_all = input_type

                for in_freq, out_freq in zip(input_freq, output_freq):
                    if in_freq not in _freq_data[neuron_name]:
                        _freq_data[neuron_name][in_freq] = dict()

                    _freq_data[neuron_name][in_freq][num_inputs] = out_freq
                    _all_input_freq.append(in_freq)

                _all_num_inputs.append(num_inputs)

        # Extract the relevant data, make sure it is sorted in right order also
        input_freq_list = np.array(sorted(list(set(_all_input_freq))))
        num_input_list = np.array(sorted(list(set(_all_num_inputs))))

        freq_data = dict()
        for neuron_name in frequency_data:
            freq_data[neuron_name] = dict()
            for input_freq in input_freq_list:
                num_in_list = []  # Number of inputs
                out_f_list = []
                for num_inputs in num_input_list:
                    if input_freq in _freq_data[neuron_name] and \
                            num_inputs in _freq_data[neuron_name][input_freq]:
                        num_in_list.append(num_inputs)
                        out_f_list.append(_freq_data[neuron_name][input_freq][num_inputs])

                freq_data[neuron_name][input_freq] = np.array(num_in_list), np.array(out_f_list)

        for neuron_name in frequency_data:
            fig, ax = plt.subplots()
            legend_text = []

            cmap = plt.get_cmap("tab20", len(freq_data[neuron_name]))
            ax.set_prop_cycle('color', [cmap(i) for i in range(0, len(freq_data[neuron_name]))])

            for input_freq in freq_data[neuron_name]:
                num_input, output_freq = freq_data[neuron_name][input_freq]
                legend_text.append(f"{input_freq} Hz input")
                ax.plot(num_input, output_freq)

            plt.title(f"{neuron_name} receiving {input_type_all} input")
            plt.xlabel("Number of synapses")
            plt.ylabel("Spikes/s")
            ax.legend(legend_text)

            if show_plots:
                plt.ion()
                plt.show()
                plt.pause(0.001)

            # fig_name = os.path.join(self.network_path, "figures", f"input-scaling-ninputs-{neuron_name}-{input_type_name}.pdf")
            fig_name = os.path.join(self.network_path, "figures",
                                    f"input-scaling-ninputs-{neuron_name}-{input_type_name}.png")
            if not os.path.exists(os.path.dirname(fig_name)):
                os.mkdir(os.path.dirname(fig_name))

            plt.savefig(fig_name, dpi=300)

            if not show_plots:
                plt.close()

    def plot_verify_frequency_distribution(self, input_type="cortical"):

        network_info, input_config, input_data, neuron_id_lookup, neuron_name_list, \
            spike_data, volt, time, depolarisation_blocks = self.load_data_helper()

        # First find out what time ranges we need to look at, and what the input frequency is for those
        neuron_type = list(input_config.keys())
        assert len(neuron_type) == 1, f"Plot only supports one neuron type at a time, neuron_type = {neuron_type}"
        neuron_type = neuron_type[0]

        assert input_type in input_config[neuron_type], f"{input_type} not in input_config: {input_config}"
        start_times = np.array(input_config[neuron_type][input_type]["start"])
        end_times = np.array(input_config[neuron_type][input_type]["end"])
        input_freq = np.array(input_config[neuron_type][input_type]["frequency"])
        output_freq_list = []

        #import pdb
        #pdb.set_trace()

        depol_blocked_freqs = []
        depol_blocked_lookup = dict()

        for neuron_id in sorted(spike_data.keys()):
            out_freq = []
            for start_t, end_t, in_freq in zip(start_times, end_times, input_freq):
                n_spikes = np.sum(np.logical_and(start_t <= spike_data[neuron_id], spike_data[neuron_id] < end_t))
                # import pdb
                # pdb.set_trace()
                f = n_spikes / (end_t - start_t)
                out_freq.append(f)

                if self.is_depolarisation_blocked(neuron_id=neuron_id, time_range=(start_t, end_t),
                                                  depolarisation_blocks=depolarisation_blocks):
                    depol_blocked_freqs.append((in_freq, f))

                    if in_freq not in depol_blocked_lookup:
                        depol_blocked_lookup[in_freq] = [f]
                    else:
                        depol_blocked_lookup[in_freq].append(f)

            output_freq_list.append(out_freq)

        output_freq = np.array(output_freq_list).T

        if len(depol_blocked_freqs) > 0:
            plt.figure()
            plt.plot(input_freq, output_freq, 'k')

            depol_in_freq, depol_out_freq = zip(*depol_blocked_freqs)
            plt.plot(depol_in_freq, depol_out_freq, 'r*')

            plt.xlabel("Input frequency")
            plt.ylabel("Output frequency")
            plt.ion()
            plt.show()

        one_mat = np.ones((2,))

        bad_idx = set()

        fig, ax = plt.subplots(len(input_freq), 1, figsize=(6, 12))

        for idx, (in_freq, out_freq) in enumerate(zip(input_freq, output_freq)):

            # ax[idx].hist(out_freq, bins=50)
            counts, bins = np.histogram(out_freq, bins=50)
            ax[idx].stairs(counts, bins, color="black")

            if in_freq in depol_blocked_lookup:
                depol_counts, depol_bins = np.histogram(depol_blocked_lookup[in_freq], bins=bins)
                ax[idx].stairs(depol_counts, depol_bins, color="red")

            yl = ax[idx].get_ylim()
            mean_freq = np.mean(out_freq)
            std_freq = np.std(out_freq)

            ax[idx].plot(one_mat*mean_freq, yl, 'k-')
            ax[idx].plot(one_mat*(mean_freq + 2*std_freq), yl, 'k--')
            ax[idx].plot(one_mat*(mean_freq - 2*std_freq), yl, 'k--')
            ax[idx].plot([in_freq, in_freq], yl, 'b-')

            bad_idx = bad_idx.union(set(np.where(np.logical_or(out_freq < mean_freq - 2*std_freq,
                                                               out_freq > mean_freq + 2*std_freq))[0]))

            ax[idx].set_xlabel("Spiking frequency (Hz)")

        plt.ion()
        plt.show()

        for idx in bad_idx:
            bad_neuron = network_info.data["neurons"][idx]
            print(f"Frequency outliers: {bad_neuron['name']} ({idx}) -- {output_freq[:, idx]} Hz\n{bad_neuron['neuron_path']}")

    def get_neuron_info(self, neuron_path):

        neuron_path = snudda_parse_path(neuron_path, self.snudda_data)

        neuron_info = collections.OrderedDict()

        # Find neuron morphology swc file, obs currently assume lowercase(!)
        neuron_morph = self.init_helper.get_morphologies(neuron_path)

        parameter_file = os.path.join(neuron_path, "parameters.json")
        mechanism_file = os.path.join(neuron_path, "mechanisms.json")
        modulation_file = os.path.join(neuron_path, "modulation.json")  # Optional
        meta_file = os.path.join(neuron_path, "meta.json")

        # Check if empty neuron_morph_list, or if more than one morphology
        assert os.path.isfile(parameter_file), f"Missing parameter file {parameter_file}"
        assert os.path.isfile(mechanism_file), f"Missing mechanism file {mechanism_file}"

        # TODO: We need to have neuron_info["neuron_path"][name_of_neuron] = neuron_path
        # The below line is wrong, but it is FRIDAY evening so will fix it monday...
        neuron_info["neuron_path"] = snudda_simplify_path(neuron_path, self.snudda_data)

        neuron_info["morphology"] = snudda_simplify_path(neuron_morph, self.snudda_data)
        neuron_info["parameters"] = snudda_simplify_path(parameter_file, self.snudda_data)
        neuron_info["mechanisms"] = snudda_simplify_path(mechanism_file, self.snudda_data)
        neuron_info["meta"] = snudda_simplify_path(meta_file, self.snudda_data)

        # Modulation file is optional
        if os.path.isfile(modulation_file):
            neuron_info["modulation"] = snudda_simplify_path(modulation_file, self.snudda_data)

        return neuron_info

    # This loops through all single neuron directories in neurons_path
    # in preparation of writing a network config file

    def gather_all_neurons(self, neuron_types=None, all_combinations=True):
        all_neurons = collections.OrderedDict()

        assert snudda_isdir(self.neurons_path, self.snudda_data), \
            f"Neurons directory {self.neurons_path} does not exist."

        neuron_type_dir = [d for d in glob.glob(os.path.join(snudda_parse_path(self.neurons_path, self.snudda_data),
                                                             '*'))
                           if snudda_isdir(d, self.snudda_data)]

        self.neuron_types = []

        if neuron_types is not None:
            if type(neuron_types) == str:
                neuron_types = [neuron_types]
            neuron_types = [x.lower() for x in neuron_types]

        for ntd in neuron_type_dir:

            neuron_type = os.path.basename(os.path.normpath(ntd))

            if neuron_types is not None:
                if neuron_type.lower() not in neuron_types:
                    print(f"Skipping neuron type {neuron_type}")
                    continue

            self.neuron_types.append(neuron_type)

            neuron_dir = [d for d in glob.glob(os.path.join(ntd, '*')) if os.path.isdir(d)]
            neuron_ctr = 0

            for nd in neuron_dir:
                neuron_info = self.get_neuron_info(nd)

                if all_combinations:
                    assert "meta" in neuron_info and neuron_info["meta"], \
                        f"meta.json required for all_combinations=True. {os.path.dirname(neuron_info['parameters'])}"

                    neuron_info_combination_list = self.get_all_combinations(neuron_info)
                    for ni in neuron_info_combination_list:
                        n_name = os.path.basename(os.path.dirname(ni["parameters"]))
                        param_key = ni["parameter_key"]
                        morph_key = ni["morphology_key"]
                        short_name = n_name[:min(10, len(n_name))]
                        neuron_name = f"{neuron_type}_{short_name}_{param_key}_{morph_key}".replace("-", "_")
                        all_neurons[neuron_name] = ni
                        neuron_ctr += 1
                else:
                    neuron_name = os.path.basename(os.path.dirname(neuron_info["parameters"]))
                    neuron_ctr += 1

                    all_neurons[neuron_name] = neuron_info

            if neuron_ctr > 0:
                print(f"Found {neuron_ctr} neuron models in {ntd}")

        assert len(all_neurons) > 0, (f"No neurons selected. Did you specify an incorrect neuronType? {neuron_types}"
                                      f"\nSee skipped neurons above error message for available ones.")

        return all_neurons

    def get_all_combinations(self, neuron_info):

        assert "meta" in neuron_info and neuron_info["meta"]

        pm_list = []

        with open(snudda_parse_path(neuron_info["parameters"], self.snudda_data), "r") as pf:
            param_data = json.load(pf)

        with open(snudda_parse_path(neuron_info["meta"], self.snudda_data), "r") as mf:
            meta_data = json.load(mf)

        for p_idx, p_key in enumerate(param_data):
            assert p_key in meta_data, f"parameter key {p_key} missing in {neuron_info['meta_file']}"
            for m_idx, m_key in enumerate(meta_data[p_key]):
                ni = copy.deepcopy(neuron_info)
                ni["morphology_key"] = m_key
                ni["parameter_key"] = p_key
                ni["morphology_file"] = os.path.join(ni["morphology"], meta_data[p_key][m_key]["morphology"])
                ni["parameter_id"] = p_idx
                ni["morphology_id"] = m_idx

                pm_list.append(ni)

        return pm_list

    def has_axon(self, neuron_info):

        nm = NeuronPrototype(neuron_name="JJJ",
                             neuron_path=None,
                             snudda_data=self.snudda_data,
                             morphology_path=neuron_info["morphology"],
                             parameter_path=neuron_info["parameters"],
                             mechanism_path=neuron_info["mechanisms"],
                             virtual_neuron=False)
        nm.instantiate()
        return nm.all_have_axon()

    def create_network_config(self,
                              neurons_path=None,
                              snudda_data=None,
                              num_replicas=10,
                              random_seed=None,
                              neuron_types=None,
                              single_neuron_path=None,
                              parameter_key=None,
                              morphology_key=None,
                              modulation_key=None,
                              all_combinations=True):

        """ Create a network with num_replicas number of replicas of each neuron found in
            neuron_path (e.g. Snudda/data/neurons). Alternatively if only one neuron model
            is required, use single_neuron_path pointing to neuron
            (e.g. Snudda/data/neurons/striatum/fs/fs_model_1)
            """

        self.neurons_path = neurons_path

        if snudda_data is None:
            snudda_data = self.snudda_data

        config_def = dict()
        config_def["snudda_data"] = snudda_data
        config_def["random_seed"], self.init_rng = SnuddaInit.setup_random_seeds(random_seed)

        region_def = dict()
        vol_name = "InputTest"
        region_def[vol_name] = dict()
        region_def[vol_name]["volume"] = dict()

        region_def[vol_name]["volume"]["type"] = "mesh"
        region_def[vol_name]["volume"]["d_min"] = 15e-6
        region_def[vol_name]["volume"]["mesh_file"] = "data/mesh/InputTestMesh.obj"
        region_def[vol_name]["volume"]["num_putative_points"] = 10000
        region_def[vol_name]["connectivity"] = dict()  # Unconnected

        if single_neuron_path:
            # Override and only get one neuron
            assert type(neuron_types) == str, "neuron_types must be string if single_neuron_path set"
            neuron_def = collections.OrderedDict()
            neuron_def[neuron_types] = self.get_neuron_info(single_neuron_path)

            if parameter_key:
                neuron_def[neuron_types]["parameter_key"] = parameter_key

            if morphology_key:
                neuron_def[neuron_types]["morphology_key"] = morphology_key

            if modulation_key:
                neuron_def[neuron_types]["modulation_key"] = modulation_key

        else:
            neuron_def = self.gather_all_neurons(neuron_types=neuron_types, all_combinations=all_combinations)

        # Just generate a set of points
        region_def[vol_name]["volume"]["n_putative_points"] = max(len(neuron_def.keys())*5, 10000)

        fake_axon_density = ["r", "1", 10e-6]

        for n in neuron_def.keys():
            neuron_def[n]["num_neurons"] = num_replicas
            neuron_def[n]["volume_id"] = vol_name
            neuron_def[n]["rotation_mode"] = "random"
            neuron_def[n]["hoc"] = None

            if isinstance(neuron_def[n]["neuron_path"], str):
                neuron_def[n]["neuron_path"] = {n: neuron_def[n]["neuron_path"]}

            if not self.has_axon(neuron_def[n]):
                print(f"One or more of morphologies {neuron_def[n]['morphology']} has no axon, faking it.")
                # We will have no connections in this test network, so add empty density
                neuron_def[n]["axon_density"] = fake_axon_density

        region_def[vol_name]["neurons"] = neuron_def

        config_def["regions"] = region_def

        return config_def

    def load_network_info(self, network_file):

        self.neuron_info = SnuddaLoad(network_file).data["neurons"]

    def collect_neurons(self):

        if not self.neuron_info:
            self.load_network_info(network_file=self.network_file)

        neuron_id = np.array([x["neuron_id"] for x in self.neuron_info])
        neuron_type = np.array([x["type"] for x in self.neuron_info])

        neuron_sets = dict()

        for nt in set(neuron_type):
            idx = np.where(neuron_type == nt)
            neuron_id_list = np.sort(neuron_id[idx])

            neuron_sets[nt] = neuron_id_list

        return neuron_sets

    # synapse_density and synapse_conductance can be either values (then same for all neurons)
    # or dictionaries, with neuron_type as key.

    def create_input_config(self,
                            input_config_file,
                            input_type,
                            n_input_min, n_input_max, num_input_steps,
                            input_frequency,
                            input_correlation,
                            synapse_density,
                            synapse_conductance,
                            synapse_parameter_file,
                            cluster_size,
                            cluster_spread,
                            input_duration=10.0):

        # assert n_input_min > 0, "No point using n_input_min=0, please instead use input_frequency 0."

        neuron_sets = self.collect_neurons()
        n_inputs = dict()

        for neuron_type in neuron_sets:

            # For each neuron model we will have num_replicas copies (see other part of code), and this
            # will determine how many steps we have between n_input_min and n_input_max

            neuron_id_list = neuron_sets[neuron_type]
            n_unique = len(neuron_id_list) / num_input_steps

            assert n_unique % 1 == 0, f"Internal error, every model version should exist {num_input_steps} times"

            num_range = np.linspace(n_input_min, n_input_max, num=num_input_steps).astype(int).reshape((1, num_input_steps)).repeat(int(n_unique), axis=0).flatten()

            for neuron_id, num_input in zip(neuron_id_list, num_range):

                if type(synapse_density) == dict:
                    sd = synapse_density[neuron_type]
                else:
                    sd = synapse_density

                if type(synapse_parameter_file) == dict:
                    if neuron_type in synapse_parameter_file:
                        spf = synapse_parameter_file[neuron_type]
                    else:
                        spf = None
                        print(f"No parameter file for {neuron_type} {input_type} input")
                else:
                    spf = synapse_parameter_file

                if type(synapse_conductance) == dict:
                    sc = synapse_conductance[neuron_type]
                else:
                    sc = synapse_conductance

                self.add_input(input_target=neuron_id,
                               input_type=input_type,
                               input_frequency=input_frequency,
                               input_correlation=input_correlation,
                               input_duration=input_duration,
                               input_density=sd,
                               num_input=num_input,
                               input_conductance=sc,
                               synapse_parameter_file=spf,
                               cluster_size=cluster_size,
                               cluster_spread=cluster_spread)

        with open(input_config_file, "w") as f:
            json.dump(self.input_info, f, indent=4, cls=NumpyEncoder)

    def add_input(self, input_target, input_type, input_frequency, input_correlation,
                  input_duration,
                  input_density, num_input, input_conductance,
                  cluster_size, cluster_spread,
                  synapse_parameter_file):

        if type(input_target) != str:
            input_target = str(input_target)

        if not self.input_info:
            self.input_info = collections.OrderedDict()

        assert type(input_frequency) == list, f"add_input: input_freq should be a list: {input_frequency}"
        n_steps = len(input_frequency)
        input_start = input_duration * np.arange(0, n_steps)
        input_end = input_duration * np.arange(1, n_steps + 1)

        if input_target not in self.input_info:
            # By only creating dictionary if it does not exist, we can have multiple inputs
            self.input_info[input_target] = collections.OrderedDict()

        self.input_info[input_target][input_type] = collections.OrderedDict()

        self.input_info[input_target][input_type]["generator"] = "poisson"
        self.input_info[input_target][input_type]["type"] = "AMPA_NMDA"
        self.input_info[input_target][input_type]["synapse_density"] = input_density
        self.input_info[input_target][input_type]["num_inputs"] = num_input
        self.input_info[input_target][input_type]["frequency"] = input_frequency
        if input_correlation is not None:
            self.input_info[input_target][input_type]["correlation"] = input_correlation
        self.input_info[input_target][input_type]["start"] = input_start
        self.input_info[input_target][input_type]["end"] = input_end
        self.input_info[input_target][input_type]["population_unit_correlation"] = 0.0
        self.input_info[input_target][input_type]["jitter"] = 0.0
        self.input_info[input_target][input_type]["conductance"] = input_conductance
        self.input_info[input_target][input_type]["mod_file"] = "tmGlut"
        if synapse_parameter_file is not None:
            self.input_info[input_target][input_type]["parameter_file"] = synapse_parameter_file

        if cluster_size is not None:
            self.input_info[input_target][input_type]["cluster_size"] = cluster_size

        if cluster_spread is not None:
            self.input_info[input_target][input_type]["cluster_spread"] = cluster_spread

    def plot_generated_input(self, num_bins=50):
        # This function just checks that we have reasonable spikes generated

        input_spike_data = h5py.File(self.input_spikes_file, 'r')
        network_data = h5py.File(self.network_file, 'r')

        neuron_type = np.array([x.decode().split("_")[0].lower() for x in network_data["network/neurons/name"]])
        neuron_id = np.array([x for x in network_data["network/neurons/neuron_id"]])

        # Plot the input spikes
        for nt in set(neuron_type):
            neuron_idx = np.where(neuron_type == nt)
            distance_to_soma = dict()

            if len(neuron_idx) == 0:
                continue

            fig, ax = plt.subplots()

            for nid in neuron_id[neuron_idx]:
                for input_type in input_spike_data["input"][str(nid)]:
                    spikes = input_spike_data["input"][str(nid)][input_type]["spikes"][:].ravel()
                    spikes = spikes[spikes >= 0]  # Negative -1 is filler values, remove them.
                    ax.hist(spikes, num_bins, histtype="step")

                    if input_type not in distance_to_soma:
                        distance_to_soma[input_type] = []

                    distance_to_soma[input_type].append(
                        input_spike_data["input"][str(nid)][input_type]["distance_to_soma"][:])

            plt.title(f"Input to {nt}")
            plt.xlabel("Time (s)")
            plt.ylabel("Count")
            plt.ion()
            plt.show()
            plt.pause(0.001)

            fig_dir = os.path.join(self.network_path, "figures")
            if not os.path.isdir(fig_dir):
                os.mkdir(fig_dir)

            fig_name1 = os.path.join(self.network_path, "figures", f"{nt}-binned-input.pdf")
            plt.savefig(fig_name1)

            fig2, ax2 = plt.subplots()
            leg = []
            for input_type in distance_to_soma:
                leg.append(input_type)
                ax2.hist(np.concatenate(distance_to_soma[input_type]) * 1e6, 50, histtype="step")

            plt.legend(leg)
            plt.title(f"Synapses onto {nt}")
            plt.xlabel(f"Distance to soma (micrometers)")
            plt.ylabel(f"Count")
            plt.show()
            plt.pause(0.001)

            fig_name2 = os.path.join(self.network_path, "figures", f"{nt}-input-distance-to-soma.pdf")
            plt.savefig(fig_name2)

        input_spike_data.close()
        network_data.close()

    def simulate(self, mech_dir=None, sample_dt=0.005):

        if self.input_seed_list is None:
            self.simulate_helper(mech_dir=mech_dir, sample_dt=sample_dt)
        else:
            for ctr, (input_file, output_file) in enumerate(zip(self.input_spikes_file, self.output_file)):
                print(f"Iteration: {ctr+1}/{len(self.input_spikes_file)}")
                print(f"Input file: {input_file}\nOutput file: {output_file}")
                self.simulate_helper(mech_dir=mech_dir, sample_dt=sample_dt,
                                     input_spikes_file=input_file, output_file=output_file)

    def simulate_helper(self, mech_dir=None, sample_dt=0.01, input_spikes_file=None, output_file=None):

        if input_spikes_file is None:
            input_spikes_file = self.input_spikes_file

        if output_file is None:
            output_file = self.output_file

        from snudda.core import Snudda
        Snudda.compile_mechanisms(mech_dir=mech_dir)

        # Get info so we can set max_time correctly
        self.read_tuning_info()

        from neuron import h  # , gui
        start = timeit.default_timer()

        pc = h.ParallelContext()

        sim = SnuddaSimulate(network_file=self.network_file,
                             input_file=input_spikes_file,
                             output_file=output_file,
                             sample_dt=sample_dt)
        sim.setup()
        sim.add_external_input()
        sim.check_memory_status()

        # sim.add_volt_recording()
        sim.add_volt_recording_soma()

        t_sim = self.max_time * 1000  # Convert from s to ms for Neuron simulator

        sim.check_memory_status()
        print(f"Running simulation for {t_sim} ms.")
        sim.run(t_sim)  # In milliseconds

        print("Simulation done, saving output")
        sim.write_output()

        stop = timeit.default_timer()
        if sim.pc.id() == 0:
            print(f"Program run time: {stop - start:.1f}s")

        print("About to clear memory")
        sim.check_memory_status()
        sim.clear_neuron()
        print("Memory cleared")
        sim.check_memory_status()
        sim = None

    def read_tuning_info(self):
        tuning_info_file = os.path.join(self.network_path, "tuning-info.json")

        if not os.path.exists(tuning_info_file):
            print("No tuning info file exists.")
            return

        try:
            with open(tuning_info_file, 'rt') as f:
                tuning_meta_data = json.load(f, object_pairs_hook=collections.OrderedDict)

            # max_time is the important one, we want to make sure we simulate correct duration without having the user
            # provide the parameter twice
            self.max_time = tuning_meta_data["max_time"]
            self.input_duration = tuning_meta_data["input_duration"]
            self.frequency_range = tuning_meta_data["frequency_range"]
            self.neurons_path = tuning_meta_data["neurons_directory"]
        except:
            print(f"Failed to read {tuning_info_file}")

    def write_tuning_info(self):
        tuning_meta_data = collections.OrderedDict()
        tuning_meta_data["input_duration"] = self.input_duration
        tuning_meta_data["max_time"] = self.max_time
        tuning_meta_data["frequency_range"] = self.frequency_range
        tuning_meta_data["neurons_directory"] = self.neurons_path

        tuning_info_file = os.path.join(self.network_path, "tuning-info.json")
        with open(tuning_info_file, "wt") as f:
            json.dump(tuning_meta_data, f, indent=4, cls=NumpyEncoder)


if __name__ == "__main__":

    if '-python' in sys.argv:
        print("Network_simulate.py called through nrniv, fixing arguments")
        pythonidx = sys.argv.index('-python')
        if len(sys.argv) > pythonidx:
            sys.argv = sys.argv[pythonidx + 1:]

    from argparse import ArgumentParser, RawTextHelpFormatter

    parser = ArgumentParser("Input Scaling", formatter_class=RawTextHelpFormatter)
    parser.add_argument("action", choices=["setup", "simulate", "analyse"], help="Action to run.")
    parser.add_argument("networkPath", help="Network path")
    parser.add_argument("--neurons", help="Neurons path")
    parser.add_argument("--mechDir", type=str, help="Path to mechanisms", default=None)
    parser.add_argument("--input_type", help="Type of external input",
                        choices=["thalamic", "cortical", "corticothalamic"],
                        default="thalamic")  # only use corticalthalamic in analyse
    parser.add_argument("--numInputSteps", type=int, help="Number of steps for number of inputs to neurons",
                        default=10)
    parser.add_argument("--numInputMin", type=int, help="Minimum number of synaptic inputs of input_type", default=100)
    parser.add_argument("--numInputMax", type=int, help="Maximum number of synaptic inputs of input_type", default=1000)
    parser.add_argument("--inputDuration", type=float, default=10.0,
                        help="Duration of each frequency test, longer need for irregularly firing neurons")
    parser.add_argument("--inputFrequency", type=str, default="[0,1,2,5]",
                        help="Input frequency, float or list of floats")
    parser.add_argument("--neuronType", default=None, type=str,
                        help="Optional, if only we want to simulate one neuron type, eg. FS")
    parser.add_argument("--singleNeuronType", default=None, type=str,
                        help="Optional, if only we want to simulate one neuron subtype, eg. FS_1")
    parser.add_argument("--meta_input", action="store_true", default=False)
    parser.add_argument("--seed_list", type=str, default=None)
    parser.add_argument("--no_downsampling", action="store_true")

    args = parser.parse_args()

    # TODO: Let the user choose input type, duration for each "run", frequency range, number of input range

    if args.seed_list is not None:
        seed_list = ast.literal_eval(args.seed_list)
    else:
        seed_list = None

    input_scaling = InputTuning(args.networkPath, input_seed_list=seed_list)

    if args.action == "setup":
        input_frequency = ast.literal_eval(args.inputFrequency)
        if type(input_frequency) != list:
            input_frequency = np.array(list(input_frequency))

        input_scaling.setup_network(neurons_path=args.neurons,
                                    num_replicas=args.numInputSteps,
                                    neuron_types=args.neuronType,
                                    single_neuron_path=args.singleNeuronType)
        input_scaling.setup_input(input_type=args.input_type,
                                  num_input_min=args.numInputMin,
                                  num_input_max=args.numInputMax,
                                  num_replicas=args.numInputSteps,
                                  input_duration=args.inputDuration,
                                  input_frequency_range=input_frequency,
                                  use_meta_input=args.no_meta_input)

        print("Tip, to run in parallel on your local machine use: "
              "mpiexec -n 4 python3 tuning/input_tuning.py simulate <yournetworkhere>")

    if args.action == "setup_background":
        input_frequency = ast.literal_eval(args.inputFrequency)

        if type(input_frequency) != list:
            input_frequency = np.array(list(input_frequency))

        if len(input_frequency) != 1:
            raise ValueError("input_frequency must only be one value when doing setup_background")

        input_scaling.setup_network(neurons_path=args.neurons,
                                    num_replicas=args.numInputSteps,
                                    neuron_types=args.neuronType,
                                    single_neuron_path=args.singleNeuronType)

        print(f"Setting up background input, will do cortical and thalamic background 50-50 at {input_frequency}")

        input_scaling.setup_background_input(input_types=["cortical_background", "thalamic_background"],
                                             input_density=["1.15*0.05/(1+exp(-(d-30e-6)/5e-6))", "0.05*exp(-d/200e-6)"],
                                             input_fraction=[0.5, 0.5],
                                             num_input_min=args.numInputMin,
                                             num_input_max=args.numInputMax,
                                             input_duration=args.inputDuration,
                                             input_frequency=[input_frequency, input_frequency])

    elif args.action == "simulate":
        print("Run simulation...")
        print("Tip, to run in parallel on your local machine use: "
              "mpiexec -n 4 python3 tuning/input_tuning.py simulate <yournetworkhere>")

        if args.no_downsampling:
            sample_dt = None
        else:
            sample_dt = 0.01

        input_scaling.simulate(mech_dir=args.mechDir, sample_dt=sample_dt)

    elif args.action == "analyse":
        # input_scaling.plot_generated_input()
        input_scaling.analyse_results(input_type=args.input_type)

    elif args.action == "analyse_background":
        input_scaling.analyse_background()

    else:
        print(f"Unknown action {args.action}")

    # python3 input_tuning/input_tuning.py setup networks/input_scaling_v1/ data/neurons/striatum/
    # mpiexec -n 4 python3 input_tuning/input_tuning.py simulate networks/input_scaling_v1/ < input.txt &> output-tuning.txt &

    #



import os
import json
import numpy as np

from snudda.utils.load import SnuddaLoad
from snudda.utils.load_network_simulation import SnuddaLoadNetworkSimulation
from snudda.utils.export_connection_matrix import SnuddaExportConnectionMatrix
from collections import OrderedDict

import matplotlib.pyplot as plt


class SnuddaAnalyseTopologyActivity:

    def __init__(self):

        self.simulation_data = dict()
        self.mapping_list = dict()
        self.mapping_dictionary = dict()

    def load_simulation_data(self, data_key, simulation_output=None):
        self.simulation_data[data_key] = SnuddaLoadNetworkSimulation(network_simulation_output_file=simulation_output)
        self.load_mapping_file(data_key)

    def load_mapping_file(self, data_key):

        network_file = SnuddaLoad.to_str(self.simulation_data[data_key].network_simulation_file["meta_data"]["network_file"][()])
        mapping_file = f"{network_file}-remapping.txt"

        self.mapping_list[data_key] = np.genfromtxt(mapping_file, delimiter=',', dtype=int)
        self.mapping_dictionary[data_key] = OrderedDict()

        for row in self.mapping_list[data_key]:
            self.mapping_dictionary[data_key][row[0]] = row[1]

    def check_same_neurons(self, data_key_a, data_key_b):

        sim_a = self.simulation_data[data_key_a]
        sim_b = self.simulation_data[data_key_b]

        is_same = len(list(sim_a.iter_neuron_id())) == len(list(sim_b.iter_neuron_id()))

        for nid_A, nid_B in zip(sim_a.iter_neuron_id(), sim_b.iter_neuron_id()):
            is_same = is_same and nid_A == nid_B
            is_same = is_same and sim_a.get_neuron_keys(nid_A) == sim_b.get_neuron_keys(nid_B)

        return is_same

    def match_closest_spikes(self, spike_train_a, spike_train_b):

        # This function assumes there are spike in both spike trains
        t_diff = (np.kron(np.ones((spike_train_a.size, 1)), spike_train_b)
                  - np.kron(spike_train_a.reshape(spike_train_a.size, 1), np.ones(spike_train_b.shape)))
        min_pos_a = np.argmin(np.abs(t_diff), axis=1)
        min_pos_b = np.argmin(np.abs(t_diff), axis=0)

        t_min_diff_a = [t_diff[m[0], m[1]] for m in zip(range(len(min_pos_b)), min_pos_a)]
        t_min_diff_b = [-t_diff[m[0], m[1]] for m in zip(min_pos_b, range(len(min_pos_a)))]

        return np.array(t_min_diff_a), np.array(t_min_diff_b)

    def match_order_spikes(self, spike_train_a, spike_train_b):

        n_compare = min(spike_train_a.size, spike_train_b.size)
        return spike_train_b[:n_compare] - spike_train_a[:n_compare], \
            spike_train_a[:n_compare] - spike_train_b[:n_compare]

    def match_closest_unique(self, spike_train_a, spike_train_b, delta_t):

        dt_list_a = np.full(spike_train_a.shape, np.nan, dtype=float)
        dt_list_b = np.full(spike_train_b.shape, np.nan, dtype=float)

        i_b = 0

        for i_a, t_a in enumerate(spike_train_a):

            while i_b < len(spike_train_b) and spike_train_b[i_b] - t_a < delta_t:
                if spike_train_b[i_b] - t_a < -delta_t:
                    i_b += 1
                else:
                    dt_list_a[i_a] = spike_train_b[i_b] - t_a
                    i_b += 1
                    break

        i_a = 0

        for i_b, t_b in enumerate(spike_train_b):

            while i_a < len(spike_train_a) and spike_train_a[i_a] - t_b < delta_t:
                if spike_train_a[i_a] - t_b < -delta_t:
                    i_a += 1
                else:
                    dt_list_b[i_b] = spike_train_a[i_a] - t_b
                    i_a += 1
                    break

        return dt_list_a, dt_list_b

    def get_spike_triggered_deltas(self, spike_times_a, spike_times_b):

        # The idea here is that for each spike in spike_train_a, we want to find the delta_t to all the spikes
        # following in spike_train_b (but before the next spike in spike_train_a). Sort of like a JPSTH.

        spike_dt = np.full(spike_times_b.shape, np.nan)

        idx_a = len(spike_times_a) - 1

        for idx_b, t_b in zip(np.flip(np.arange(0, len(spike_times_b))), np.flip(spike_times_b)):
            while idx_a >= 0 and spike_times_a[idx_a] > t_b:
                idx_a -= 1

            spike_dt[idx_b] = t_b - spike_times_a[idx_a] if idx_a >= 0 else np.nan

        return spike_dt

    def get_triggered_spikes(self, spike_times, trigger_times, duration, subtract_trigger=True):

        start_idx = 0

        for trigger_t in trigger_times:
            while start_idx < len(spike_times) and spike_times[start_idx] < trigger_t:
                start_idx += 1

            end_idx = start_idx

            while end_idx < len(spike_times) and spike_times[end_idx] <= trigger_t + duration:
                end_idx += 1

            if start_idx < len(spike_times):
                if subtract_trigger:
                    yield spike_times[start_idx:end_idx] - trigger_t
                else:
                    yield spike_times[start_idx:end_idx]

    def get_jpsth(self, spike_times_a, spike_times_b, trigger_times, duration=None, bin_size=2e-3):

        if duration is None and len(trigger_times) > 1:
            duration = trigger_times[1] - trigger_times[0]

        jpsth = np.zeros((bin_size, bin_size), dtype=int)

        for spikes_a, spikes_b in zip(self.get_triggered_spikes(spike_times=spike_times_a,
                                                                trigger_times=trigger_times,
                                                                duration=duration),
                                      self.get_triggered_spikes(spike_times=spike_times_b,
                                                                trigger_times=trigger_times,
                                                                duration=duration)):

            for idx_a in np.floor(spikes_a / bin_size):
                for idx_b in np.floor(spikes_b / bin_size):
                    jpsth[idx_a, idx_b] += 1

        return jpsth

    def plot_jpsth(self, spike_times_a, spike_times_b, trigger_times, duration=None, bin_size=2e-3, title=None):

        if duration is None and len(trigger_times) > 1:
            duration = trigger_times[1] - trigger_times[0]

        spike_combos = []

        for spikes_a, spikes_b in zip(self.get_triggered_spikes(spike_times=spike_times_a,
                                                                trigger_times=trigger_times,
                                                                duration=duration),
                                      self.get_triggered_spikes(spike_times=spike_times_b,
                                                                trigger_times=trigger_times,
                                                                duration=duration)):
            for t_a in spikes_a:
                for t_b in spikes_b:
                    spike_combos.append((t_a, t_b))

        x, y = zip(*spike_combos)

        plt.figure()
        plt.hist2d(x=np.array(x), y=np.array(y),
                   bins=int(np.ceil(duration/bin_size)+1), range=[[0, duration], [0, duration]],
                   cmap=plt.get_cmap("Reds"))
        plt.plot([0, duration], [0, duration], color="k", linestyle="dotted")
        plt.colorbar()
        plt.title(title)
        plt.xlabel("Ablated: time (s)")
        plt.ylabel("Fully ablated: time (s)")
        plt.ion()
        plt.show()

    def plot_jpsth_all(self, data_key_a, data_key_b, trigger_times, duration=None, bin_size=2e-3, time_range=None):

        if duration is None and len(trigger_times) > 1:
            duration = trigger_times[1] - trigger_times[0]

        assert self.check_same_neurons(data_key_a, data_key_b), f"data_keys have different neurons in the network"

        neuron_names = self.simulation_data[data_key_a].get_neuron_name()

        # Match spikes against each other, compute change...
        sim_a = self.simulation_data[data_key_a]
        sim_b = self.simulation_data[data_key_b]

        spikes_a = sim_a.get_spikes()
        spikes_b = sim_b.get_spikes()

        for neuron_id in spikes_a.keys():
            s_a = self.filter_times(spikes_a[neuron_id].flatten(), time_range)
            s_b = self.filter_times(spikes_b[neuron_id].flatten(), time_range)

            if s_a.size > 0 and s_b.size > 0:
                self.plot_jpsth(spike_times_a=s_a, spike_times_b=s_b,
                                trigger_times=trigger_times, duration=duration, bin_size=bin_size,
                                title=f"{neuron_names[neuron_id]} ({neuron_id})")

    def get_spike_deltas(self, data_key_a, data_key_b, matching_method, delta_t=5e-3, time_range=None):

        """
        Args:
            data_key_a : Data key for first dataset
            data_key_b : Data key for second dataset
            matching_method : Method to use for spike matching "closest", "order", "closestunique", "spike_triggered"
            delta_t : Optional, used by "closestunique" method (default 5e-3s)

        """

        # Check that the neurons compared are the same (by verifying parameter key, morphology key, modulation key)
        assert self.check_same_neurons(data_key_a, data_key_b), f"data_keys have different neurons in the network"

        # Match spikes against each other, compute change...

        sim_a = self.simulation_data[data_key_a]
        sim_b = self.simulation_data[data_key_b]

        spikes_a = sim_a.get_spikes()
        spikes_b = sim_b.get_spikes()

        spike_time_difference = dict()

        for neuron_id in spikes_a.keys():
            s_a = self.filter_times(spikes_a[neuron_id].flatten(), time_range)
            s_b = self.filter_times(spikes_b[neuron_id].flatten(), time_range)

            if s_a.size > 0 and s_b.size > 0:

                if matching_method == "closest":
                    spike_time_difference[neuron_id] = self.match_closest_spikes(s_a, s_b)
                elif matching_method == "order":
                    n_compare = min(s_a.size, s_b.size)
                    spike_time_difference[neuron_id] = self.match_order_spikes(s_a, s_b)
                elif matching_method == "closestunique":
                    spike_time_difference[neuron_id] = self.match_closest_unique(s_a, s_b, delta_t=delta_t)
                elif matching_method == "spike_triggered":
                    spike_time_difference[neuron_id] = self.get_spike_triggered_deltas(s_a, s_b), np.array([])
                else:
                    assert False, f"Unknown matching_method={matching_method}, " \
                                  f"please use ('closest', 'order', 'closestunique', 'spike_triggered')"
            else:
                # At least one of the spike trains does not have any spikes
                spike_time_difference[neuron_id] = np.array([]), np.array([])

        return spike_time_difference

    def filter_times(self, spike_times, time_range=None):

        if time_range is None:
            return spike_times
        else:
            return spike_times[time_range[0] <= spike_times <= time_range[1]]

    def plot_spike_delta_histogram(self, data_key_a=None, data_key_b=None,
                                   plot_title=None, direction=0,
                                   matching_method=None,
                                   range_min=-10e-3, range_max=2e-3, bin_size=0.5e-3):

        spike_time_difference = self.get_spike_deltas(data_key_a=data_key_a, data_key_b=data_key_b,
                                                      matching_method=matching_method)
        n_bins = int(np.ceil((range_max-range_min) / bin_size)) + 1

        fig = plt.figure()
        neuron_names = self.simulation_data[data_key_a].get_neuron_name()
        plt.set_cmap("autumn")

        hist_data = []
        hist_label = []

        for nid in spike_time_difference.keys():
            hist_data.append(spike_time_difference[nid][direction])
            hist_label.append(f"{neuron_names[nid]} ({nid})")

        plt.hist(hist_data, range=(range_min, range_max), bins=n_bins,
                 label=hist_label, histtype="barstacked")

            #plt.ion()
            #plt.show()
            #import pdb
            #pdb.set_trace()

        plt.xlabel("Time difference (s)")
        plt.ylabel("Count")
        plt.title(plot_title)
        plt.legend()
        plt.ion()
        plt.show()


import os
import pickle
import re
import sys
import timeit

import numpy as np
import scipy
from numba import jit
from scipy import ndimage


class RegionMesh(object):
    """ Handles neuron placement within a 3D mesh. """

    def __init__(self, filename, d_view=None, role="master",
                 use_cache=True, pickle_version=-1, raytrace_borders=True,
                 d_min=15e-6, bin_width=1e-4,
                 logfile_name=None, log_file=None,
                 random_seed=112, verbose=False):

        raise DeprecationWarning("Old version")
        assert False

        """
        Constructor.

        Args:
            filename (str) : Path to wavefront mesh file
            d_view : ipyparallel direct view object
            role (str) : Role, ie. "master" or "worker"
            use_cache (bool) : The meshes voxel representation is cached, use the cache?
            pickle_version (int) : Which version of pickle to use? (-1 latest)
            raytrace_borders (bool) : Raytrace positions in border regions? Slower, but more accurate.
            d_min (float) : Closest distance between soma
            bin_width (float) : Mesh size in meters
            logfile_name (str) : Path to logfile
            log_file : File pointer to log file
            random_seed (int) : Random seed value
            verbose (bool) : Flag to be verbose?

        """

        self.d_view = d_view

        self.role = role
        self.workers_initialised = False
        self.cache_file_version = 1.11

        self.verbose = verbose
        self.close_log_file = False

        if log_file is not None:
            self.logfile = log_file
            self.logfile_name = log_file.name
        elif logfile_name is not None and len(logfile_name) > 0:
            self.logfile = open(logfile_name, 'wt')
            self.logfile_name = logfile_name
            self.close_log_file = True
        else:
            self.logfile = None
            self.logfile_name = None

        self.bin_width = bin_width
        self.d_min = d_min
        self.padding = max(self.bin_width, d_min)
        self.num_bins = None

        self.density_function = dict()
        self.density_voxel_sum = dict()
        self.density_total_sum = dict()
        self.placed_voxel = dict()
        self.placed_total = dict()
        self.density_voxel_n_sample = dict()
        self.density_total_n_sample = dict()

        self.last_neuron_type_added = None
        self.density_lookup = None

        self.random_seed = random_seed
        self.random_generator = np.random.default_rng(self.random_seed)

        # Set by load_mesh
        self.mesh_vec = None
        self.mesh_faces = None
        self.mesh_norm = None
        self.min_coord = None
        self.max_coord = None
        self.voxel_mask_inner = None
        self.voxel_mask_border = None
        self.point_out = None

        # Set by pre_compute
        self.mesh_u = None
        self.mesh_v = None
        self.mesh_v0 = None
        self.mesh_uv = None
        self.mesh_vv = None
        self.mesh_uu = None
        self.mesh_denom = None
        self.mesh_nrm = None

        # Used by setup_place_neurons
        self.max_rand = 10000
        self.max_neurons = 3000000
        self.max_reject = 100e6

        # Set by setup_place_neurons
        self.rand_ctr = None
        self.random_pool = None
        self.neuron_coords = None
        self.neuron_ctr = None
        self.padding_coords = None
        self.padding_ctr = None
        self.neuron_types = None
        self.neuron_type = None
        self.next_neuron_type = 1
        self.reject_ctr = None

        # Used or set by setup_voxel_list, we assume no more than 300k neurons per mm3
        self.max_neurons_per_voxel = int(np.ceil(300000 * (self.bin_width / 1e-3) ** 3))
        self.voxel_next_neuron = None
        self.voxel_neurons = None

        # Set by update_padding_mask
        self.voxel_mask_padding = None

        # This determines if we ray trace the border voxels, for finer detail
        # or not (activating this is SLOW)
        self.raytrace_borders = raytrace_borders

        if raytrace_borders:
            self.write_log("Ray tracing points in border voxels, this is slow.")
            rt_str = "-RTB"
        else:
            rt_str = ""

        # binWidth 5e-4 (94s) --> 10.8 % border voxels
        # binWidth 2.5e-4 (577) --> 6.2 % border voxels
        # binWidth 1e-4 (8090s) --> 2.7 % border voxels
        # binWidth 0.5e-4 (??? s) --> ?? % border voxels

        assert os.path.exists(filename), f"Mesh file {filename} missing"

        self.filename = filename
        self.cache_file = f"{filename}-{int(1e6 * self.bin_width)}{rt_str}-cache.pickle"
        self.pickle_version = pickle_version

        self.debug_flag = False

        cache_loaded = False

        if use_cache and self.cache_exist():
            try:
                self.load_cache()
                cache_loaded = True
            except:
                self.write_log("Failed to load cache.")
                cache_loaded = False

        if not cache_loaded:
            self.load_mesh(filename=filename)

            t_a = timeit.default_timer()

            self.pre_compute()
            self.setup_voxel_filter()

            t_b = timeit.default_timer()

            self.write_log(f"Calculation time: {t_b - t_a} s")

        self.setup_voxel_list()

        self.setup_place_neurons(d_min=d_min)

        self.write_log(f"Inner voxel bin volume: {np.round(self.inner_voxel_volume() * 1e9, 1)} mm³")
        # !!! RUN THIS
        # self.verifyInside(100)

    ############################################################################

    def mark_borders(self):

        """ Mark border voxels in 3D mesh. """

        self.write_log("Marking borders")

        for row in self.mesh_vec:
            ic = np.floor((row - self.min_coord) / self.bin_width)
            self.voxel_mask_border[int(ic[0]), int(ic[1]), int(ic[2])] = 1

        max_num = 0

        for row in self.mesh_faces:
            coord1 = self.mesh_vec[row[0], :]
            coord2 = self.mesh_vec[row[1], :]
            coord3 = self.mesh_vec[row[2], :]

            vx = coord2 - coord1
            vy = coord3 - coord1

            dx = np.linalg.norm(coord2 - coord1)
            dy = np.linalg.norm(coord3 - coord1)

            nx = int(10 * np.ceil(dx / self.bin_width)) + 11
            ny = int(10 * np.ceil(dy / self.bin_width)) + 11

            max_num = max(max_num, max(nx, ny))

            for x_step in np.linspace(0, 1, nx):
                for y_step in np.linspace(0, 1 - x_step, ny):
                    x_point = coord1 + vx * x_step + vy * y_step
                    ic = np.floor((x_point - self.min_coord) / self.bin_width)
                    self.voxel_mask_border[int(ic[0]), int(ic[1]), int(ic[2])] = 1

        self.write_log(f"max_num = {max_num}")

    def __del__(self):
        if self.close_log_file:
            self.logfile.close()

    ############################################################################

    def setup_parallel(self, d_view=None):

        """ Setup workers for parallel execution. """

        if d_view is None:

            if self.d_view is None:
                self.write_log("Running in serial")
                return

            d_view = self.d_view

        self.write_log("Setting up parallel")

        assert self.role == "master", \
            "setup_parallel should only be called by master"

        if self.workers_initialised:
            self.write_log("setup_parallel: workers already initialised")
            return

        with d_view.sync_imports():
            from snudda.place.region_mesh import RegionMesh

        try:
            self.write_log("Setting up RegionMesh on workers")
            d_view.push({"filename": self.filename}, block=True)

            if self.logfile_name is not None:
                engine_log_file = [f"{self.logfile_name}-{x}" for x in range(0, len(d_view))]
            else:
                engine_log_file = [[] for x in range(0, len(d_view))]

            d_view.scatter('logfile_name', engine_log_file, block=True)
            d_view.push({"raytrace_borders": self.raytrace_borders,
                         "d_min": self.d_min,
                         "bin_width": self.bin_width})

            cmd_str = ("sm = RegionMesh(filename=filename, role='worker', logfile_name=logfile_name[0],"
                       "raytrace_borders=raytrace_borders, d_min=d_min, bin_width=bin_width)")

            d_view.execute(cmd_str, block=True)
            self.write_log("Worker RegionMesh setup done.")

        except Exception as e:
            import uuid
            import traceback
            tstr = traceback.format_exc()
            tmp = open(os.path.join("save", f"tmp-striatum-mesh-log-file-{uuid.uuid4()}"), 'w')
            tmp.write(f"Exception: {e}")
            tmp.write(f"Trace: {tstr}")
            tmp.close()
            self.write_log(tstr, is_error=True)
            import pdb
            pdb.set_trace()

        self.workers_initialised = True

    ############################################################################

    def cache_exist(self):

        """ Check if cache for 3D mesh exists. Returns True or False. """

        cache_flag = False

        if os.path.isfile(self.cache_file):

            obj_time = os.path.getmtime(self.filename)
            cache_time = os.path.getmtime(self.cache_file)

            if cache_time > obj_time:
                self.write_log(f"Found cache file {self.cache_file}")
                cache_flag = True
            else:
                self.write_log(f"Found old cache file ({self.cache_file}), ignoring.")
        else:
            self.write_log(f"No mesh cache file found ({self.cache_file})")

        return cache_flag

    ############################################################################

    def load_cache(self):

        """ Loading 3D mesh cache. """

        self.write_log(f"Loading cache file: {self.cache_file}")

        with open(self.cache_file, 'rb') as f:
            data = pickle.load(f)

        self.mesh_vec = data["meshVec"]
        self.mesh_faces = data["meshFaces"]
        self.mesh_norm = data["meshNorm"]
        self.min_coord = data["minCoord"]
        self.max_coord = data["maxCoord"]
        self.voxel_mask_inner = data["voxelMaskInner"]
        self.voxel_mask_border = data["voxelMaskBorder"]

        assert self.cache_file_version == data["version"], f"Old mesh cache version."

        self.point_out = self.max_coord + np.array([1e-1, 1e-2, 1e-3])
        # self.point_out = self.max_coord + 1e-2

        assert self.bin_width == data["binWidth"], \
            f"Mismatch binWidth: {self.bin_width} vs {data['binWidth']}"
        assert self.padding == data["padding"], \
            f"Mismatch padding: {self.padding} vs {data['padding']}"

        assert self.raytrace_borders == data["raytraceBorders"], \
            f"Mismatch raytraceBorders: {self.raytrace_borders} vs {data['raytraceBorders']}"

        self.num_bins = self.voxel_mask_inner.shape
        self.pre_compute()

    ############################################################################

    def save_cache(self):

        """ Save 3D mesh cache. """

        if self.role != "master":
            return

        data = dict([])
        data["meshVec"] = self.mesh_vec
        data["meshFaces"] = self.mesh_faces
        data["meshNorm"] = self.mesh_norm
        data["minCoord"] = self.min_coord
        data["maxCoord"] = self.max_coord
        data["voxelMaskInner"] = self.voxel_mask_inner
        data["voxelMaskBorder"] = self.voxel_mask_border
        data["padding"] = self.padding
        data["binWidth"] = self.bin_width
        data["raytraceBorders"] = self.raytrace_borders
        data["version"] = self.cache_file_version

        self.write_log(f"Saving mesh cache file {self.cache_file}")
        with open(self.cache_file, 'wb') as f:
            pickle.dump(data, f, self.pickle_version)

    ############################################################################

    def load_mesh(self, filename):

        """ Load 3D mesh. """

        self.filename = filename

        all_vec = []
        all_faces = []
        all_norm = []

        # https://stackoverflow.com/questions/4703390/how-to-extract-a-floating-number-from-a-string
        numeric_const_pattern = r"""
    [-+]? # optional sign
    (?:
    (?: \d* \. \d+ ) # .1 .12 .123 etc 9.1 etc 98.1 etc
    |
    (?: \d+ \.? ) # 1. 12. 123. etc 1 12 123 etc
    )
    # followed by optional exponent part if desired
    (?: [Ee] [+-]? \d+ ) ?
    """
        rx = re.compile(numeric_const_pattern, re.VERBOSE)

        with open(filename, 'rt') as f:
            for row in f:
                if row[0:2] == 'v ':
                    digits = rx.findall(row)
                    # Convert to SI units
                    all_vec.append([float(d) * 1e-6 for d in digits])

                if row[0:2] == 'f ':
                    # Only take first value of each triplet 1/?/? 2/?/? 3/?/?
                    digits = re.findall(r'f\s+(\d+)/\d*/\d*\s+(\d+)//\d*\s+(\d+)//\d*',
                                        row)
                    # Subtract one, to get python indexing
                    try:
                        all_faces.append([int(d) - 1 for d in digits[0]])
                    except:
                        self.write_log("Problem with reading digits", is_error=True)
                        self.write_log(f"{row}\nread: {digits}", is_error=True)
                        import pdb
                        pdb.set_trace()

                if row[0:2] == 'vn':
                    digits = rx.findall(row)
                    all_norm.append([float(d) for d in digits])

            self.mesh_vec = np.zeros((len(all_vec), 3))
            self.mesh_faces = np.zeros((len(all_faces), 3), dtype=int)
            self.mesh_norm = np.zeros((len(all_norm), 3))

            for ir, row in enumerate(all_vec):
                self.mesh_vec[ir, :] = row

            for ir, row in enumerate(all_faces):
                self.mesh_faces[ir, ] = row

            for ir, row in enumerate(all_norm):
                self.mesh_norm[ir, :] = row

            try:
                self.min_coord = np.min(self.mesh_vec, axis=0)
                self.max_coord = np.max(self.mesh_vec, axis=0)
            except:
                self.write_log("Problem calculating min and max coordinates")
                import pdb
                pdb.set_trace()

            # Used by ray casting when checking if another point is interior
            self.point_out = self.max_coord + np.array([1e-1, 1e-2, 1e-3])
            # self.point_out = self.max_coord + 1e-2

    ############################################################################

    def pre_compute(self):

        """ Helper function, precomputes values for raytracing. """

        i0 = self.mesh_faces[:, 0]
        i1 = self.mesh_faces[:, 1]
        i2 = self.mesh_faces[:, 2]

        self.mesh_u = self.mesh_vec[i1, :] - self.mesh_vec[i0, :]
        self.mesh_v = self.mesh_vec[i2, :] - self.mesh_vec[i0, :]

        self.mesh_v0 = self.mesh_vec[i0, :]

        self.mesh_uv = np.sum(np.multiply(self.mesh_u, self.mesh_v), axis=1)
        self.mesh_vv = np.sum(np.multiply(self.mesh_v, self.mesh_v), axis=1)
        self.mesh_uu = np.sum(np.multiply(self.mesh_u, self.mesh_u), axis=1)

        self.mesh_denom = np.multiply(self.mesh_uv, self.mesh_uv) - np.multiply(self.mesh_uu, self.mesh_vv)

        # Normal of triangle
        self.mesh_nrm = np.cross(self.mesh_u, self.mesh_v)

        # We need to normalise it
        nl = np.repeat(np.reshape(self.mesh_uv, [self.mesh_uv.shape[0], 1]), 3, axis=1)
        self.mesh_nrm = np.divide(self.mesh_nrm, nl)

    ############################################################################

    def setup_voxel_filter(self):

        """ Setup voxel filter for 3D mesh. """

        if self.role == "master":
            self.setup_parallel()

        self.min_coord = np.floor((np.min(self.mesh_vec, axis=0) - self.padding) / self.bin_width) * self.bin_width
        self.max_coord = np.ceil((np.max(self.mesh_vec, axis=0) + self.padding) / self.bin_width) * self.bin_width

        self.num_bins = np.array(np.ceil((self.max_coord - self.min_coord) / self.bin_width + 1), dtype=int)

        self.write_log(f"Voxel mask: {self.num_bins[0]} x {self.num_bins[1]} x {self.num_bins[2]}")
        self.write_log(f"min coord: {self.min_coord}, max coord {self.max_coord},"
                       f"\npadding {self.padding}, bin width {self.bin_width}")

        self.voxel_mask_inner = np.zeros(self.num_bins, dtype=bool)
        self.voxel_mask_border = np.zeros(self.num_bins, dtype=bool)

        # All voxels with a mesh point in them are "border voxels"
        # For all remaining points, check if inner or outer
        # If raytraceBorders is false, we dont raytace for border points
        # at run time, this gives a bit jagged edges, but is MUCH faster
        # when placing cells (no ray tracing then)

        if self.raytrace_borders:
            self.mark_borders()

        num_bins_total = self.num_bins[0] * self.num_bins[1] * self.num_bins[2]
        iter_ctr = 0

        # This second part is only run by the master, it calls the workers
        # to perform part of the computation

        if self.role == "master":
            # This should only be done by master

            if np.prod(self.num_bins) > 1e6:
                self.write_log(f"Calculating {np.prod(self.num_bins)} voxels. Size warning, check mesh size.",
                               force_print=True)

            if self.d_view is None:

                # No workers, do all work ourselves
                # The worker function adds a dimension (so gather works in parallel
                # case), here we just need to reshape results.
                vm_inner = self._voxel_mask_helper(range(0, self.num_bins[0]))
                self.voxel_mask_inner = np.reshape(vm_inner, self.num_bins)

            else:
                # Distribute the work to the workers
                # Randomize order, to spread work load a bit better -- order should not affect computation
                # as computation is deterministic
                all_x = np.random.permutation(np.arange(0, self.num_bins[0]))

                self.d_view.scatter("x_range", all_x, block=True)
                self.write_log("Starting parallel job")
                self.d_view.execute("inner_mask = sm._voxel_mask_helper(x_range)", block=True)
                self.write_log("Gathering results")
                inner_mask = self.d_view.gather("inner_mask", block=True)

                for m in inner_mask:
                    self.voxel_mask_inner = np.logical_or(self.voxel_mask_inner, m)

        self.write_log(f"Fraction of border voxels: "
                       f"{np.sum(self.voxel_mask_border) / np.prod(self.voxel_mask_border.shape)}")

        self.save_cache()

        if np.sum(self.voxel_mask_inner) == 0:
            self.write_log(f"!!! Warning no inner voxels in mesh, is your meshBinWidth={self.bin_width} too large?"
                           f"\nThis will prevent neurons from being placed in the volume.",
                           is_error=True)
            self.write_log(f"mesh file: {self.filename}", is_error=True)

    ############################################################################

    def check_inside(self, coords):

        """ Check if coordinates are inside 3D mesh. """

        idx = np.array(np.floor((coords - self.min_coord) / self.bin_width), dtype=int)

        if self.voxel_mask_inner[idx[0], idx[1], idx[2]]:
            # We know it is an inner voxel
            return True
        elif self.voxel_mask_border[idx[0], idx[1], idx[2]]:
            # We are in a border voxel, need to ray cast this
            return self.ray_casting(coords)
        else:
            # We are outside structure
            return False

    ############################################################################

    def _voxel_mask_helper(self, x_range):

        """ Helper function. Determine if voxels are inner our outer."""

        try:

            # Need the extra dimension at the top for "gather" work
            vm_inner = np.zeros((1, self.num_bins[0], self.num_bins[1], self.num_bins[2]), dtype=bool)

            for ix in x_range:
                self.write_log(f"Processing x = {ix}")

                for iy in range(0, self.num_bins[1]):
                    # print(f"Processing x = {ix}/{self.num_bins[0]}, y = {iy}/{self.num_bins[1]}")

                    for iz in range(0, self.num_bins[2]):

                        if not self.voxel_mask_border[ix, iy, iz]:
                            # Inner or outer point, check centre
                            xyz = np.array([self.min_coord[0] + (ix + 0.5) * self.bin_width,
                                            self.min_coord[1] + (iy + 0.5) * self.bin_width,
                                            self.min_coord[2] + (iz + 0.5) * self.bin_width])

                            vm_inner[0, ix, iy, iz] = self.ray_casting(xyz)

        except Exception as e:
            # Write error to log file to help trace it.
            import traceback
            t_str = traceback.format_exc()
            self.write_log(t_str, is_error=True)

            sys.exit(-1)

        return vm_inner

    ############################################################################

    # Cast a ray, see how many times it intersects the triangles of the structure
    # if an odd number of intersections, then it is an inner point
    #
    # Based on: https://www.erikrotteveel.com/python/three-dimensional-ray-tracing-in-python/
    # which is based on : source: http://geomalgorithms.com/a06-_intersect-2.html

    # TODO: When the line between the interior and exterior point crosses the line between two vertexes this code
    #       might incorrectly say the line is outside by considering it crosses both lines

    # TODO: Vectorise this to speed it up
    def ray_casting_OLD(self, point):

        n_tri = self.mesh_faces.shape[0]

        P = self.point_out - point
        # rn = nominator, rd = denominator
        rn = np.sum(np.multiply(self.mesh_nrm, self.mesh_v0 - point), axis=1)
        rd = np.dot(self.mesh_nrm, P)

        # r = np.divide(rn,rd)

        intersect_count = 0

        for i in range(0, n_tri):

            if rd[i] == 0:
                if rn[i] == 0:
                    # Parallel and lies in the plane
                    rI = 0.0
                else:
                    # Parallel to plane, but outside. Mark by -1 to avoid counting
                    rI = -1
            else:
                rI = rn[i] / rd[i]

            if 0 <= rI <= 1:
                # Crosses the plane, but is it within triangle?

                w = point + rI * P - self.mesh_v0[i, :]

                si = (self.mesh_uv[i] * np.inner(w, self.mesh_v[i, :])
                      - self.mesh_vv[i] * np.inner(w, self.mesh_u[i, :])) / self.mesh_denom[i]

                if si < 0 or si > 1:
                    # outside of triangle
                    continue

                ti = (self.mesh_uv[i] * np.inner(w, self.mesh_u[i, :])
                      - self.mesh_uu[i] * np.inner(w, self.mesh_v[i, :])) / self.mesh_denom[i]

                if ti < 0 or (si + ti) > 1:
                    # outside of triangle
                    continue

                # print("intersects face i = " + str(i))
                # print("si = " +str(si) + ", ti = " + str(ti))

                intersect_count += 1

        print(f"ray_casting_OLD - intersection count {intersect_count}")
        return np.mod(intersect_count, 2) == 1

    ############################################################################

    def ray_casting(self, point):

        """ Ray-casting, to determine if a point is inside or outside of mesh. """

        return RegionMesh.ray_casting_helper(point=point,
                                             self_mesh_faces=self.mesh_faces,
                                             self_mesh_nrm=self.mesh_nrm,
                                             self_mesh_v0=self.mesh_v0,
                                             self_point_out=self.point_out,
                                             self_mesh_denom=self.mesh_denom,
                                             self_mesh_uv=self.mesh_uv,
                                             self_mesh_uu=self.mesh_uu,
                                             self_mesh_vv=self.mesh_vv,
                                             self_mesh_u=self.mesh_u,
                                             self_mesh_v=self.mesh_v)

    @staticmethod
    @jit(nopython=True)
    def ray_casting_helper(point,
                           self_mesh_faces, self_mesh_nrm, self_point_out,
                           self_mesh_v0, self_mesh_denom,
                           self_mesh_uv, self_mesh_vv, self_mesh_uu, self_mesh_u, self_mesh_v):

        """
        Helper function for ray-casting, to determine if a point is inside the 3D-mesh.
        It draws a line from the point given, and a second point defined as outside.
        If that line intersects the surface of the 3D mesh an odd number of times, then the first point is inside.

        Uses values pre-computed by pre_compute function.
        """

        # print(f"Processing {point}")

        n_tri = self_mesh_faces.shape[0]

        p = self_point_out - point
        # rn = nominator, rd = denominator
        rn = np.sum(np.multiply(self_mesh_nrm, self_mesh_v0 - point), axis=1)
        rd = np.dot(self_mesh_nrm, p)

        # If rd == 0 and rn != 0 --> r = -1, parallel to plane, but outside, mark -1 to avoid counting
        # If rd == 0 and rn == 0 --> r = 0, parallel and lies in plane
        idx0 = (rd == 0)
        idx1 = np.logical_and(idx0, rn != 0)

        rn[idx1] = -1
        rd[idx0] = 1
        r = np.divide(rn, rd)

        intersect_count = 0

        idx = np.where(np.logical_and(0 <= r, r <= 1))[0]

        w = point + r.reshape(len(r), 1) * p.reshape(1, 3) - self_mesh_v0
        n_points = len(r)

        s = np.divide(np.multiply(self_mesh_uv, np.sum(np.multiply(w, self_mesh_v), axis=1).reshape(n_points, ))
                      - np.multiply(self_mesh_vv, np.sum(np.multiply(w, self_mesh_u), axis=1).reshape(n_points, )),
                      self_mesh_denom)

        t = np.divide(np.multiply(self_mesh_uv, np.sum(np.multiply(w, self_mesh_u), axis=1).reshape(n_points, ))
                      - np.multiply(self_mesh_uu, np.sum(np.multiply(w, self_mesh_v), axis=1).reshape(n_points, )),
                      self_mesh_denom)

        intersect_count = np.sum((0 <= r) * (r <= 1) * (0 <= s) * (s <= 1) * (0 <= t) * (s + t <= 1))

        return np.mod(intersect_count, 2) == 1

    ############################################################################

    def verify_inside(self, num_points=1000):

        """ Verify the check-inside method against the ray-casting method to make sure they give same results. """

        if self.role != "master":
            return

        # Use default random number generator for verification, so different each run -- more chances to catch error
        x_test = np.random.uniform(self.min_coord[0], self.max_coord[0], num_points)
        y_test = np.random.uniform(self.min_coord[1], self.max_coord[1], num_points)
        z_test = np.random.uniform(self.min_coord[2], self.max_coord[2], num_points)

        self.write_log("Verifying our method")

        for i in range(0, num_points):

            self.write_log(f"{i}/{num_points}")
            coords = np.array([x_test[i], y_test[i], z_test[i]])

            try:
                assert (self.check_inside(coords) == self.ray_casting(coords))
            except:
                self.write_log(f"Mismatch for coordinates: {coords}", is_error=True)
                self.write_log(f"Cached: {self.check_inside(coords)}", is_error=True)
                self.write_log(f"RC: {self.ray_casting(coords)}", is_error=True)
                import pdb
                pdb.set_trace()

    ############################################################################

    def setup_place_neurons(self, d_min=None):

        """
        Initialises variables for neuron placement.

        Args:
            d_min (float) : Minimal distance between neuron somas, in SI units (meters)
        """

        self.write_log("Setup place neurons")

        self.d_min = d_min

        self.rand_ctr = self.max_rand + 1

        self.random_pool = np.zeros((self.max_rand, 3))
        self.density_lookup = np.zeros((self.max_rand,))

        self.neuron_coords = np.zeros((self.max_neurons, 3))
        self.neuron_ctr = 0

        self.padding_coords = np.zeros((self.max_neurons, 3))
        self.padding_ctr = 0

        self.neuron_types = dict([])
        self.neuron_type = np.zeros((self.max_neurons,))
        self.next_neuron_type = 1

        self.reject_ctr = 0

        self.num_bins = self.voxel_mask_inner.shape

        self.update_padding_mask()
        self.update_random_pool()

        self.write_log("Setup done")

    ############################################################################

    def setup_voxel_list(self):

        """ Setup voxel list, these are voxels that needs to be checked for neurons close by. """

        self.write_log("Setup voxel list")

        self.voxel_next_neuron = np.zeros(self.num_bins, dtype=int)
        self.voxel_neurons = np.zeros((self.num_bins[0], self.num_bins[1],
                                       self.num_bins[2], self.max_neurons_per_voxel, 3))

    ############################################################################

    def update_padding_mask(self):

        """ Updates padding mask. We add neurons outside our region of interest, to avoid artificially inflating
            neuron density at the edges (which would happen without the padding region). """

        self.write_log("Update padding mask")

        # Only save padding for border voxels, and voxels nearby
        n_dist = int(np.ceil(2 * self.d_min / self.bin_width))
        s = ndimage.generate_binary_structure(3, 3)

        if np.sum(self.voxel_mask_border):
            # If we do raytracing then the border exists
            dilated_mask = scipy.ndimage.binary_dilation(self.voxel_mask_border,
                                                         structure=s,
                                                         iterations=n_dist)
        else:
            # No ray tracing, we need to create a padding region

            dilated_mask = scipy.ndimage.binary_dilation(self.voxel_mask_inner,
                                                         structure=s,
                                                         iterations=n_dist)

        self.voxel_mask_padding = np.logical_and(dilated_mask, ~self.voxel_mask_inner)

    ############################################################################

    def check_padding_zone(self, coords):

        idx = np.array(np.floor((coords - self.min_coord) / self.bin_width), dtype=int)

        return self.voxel_mask_padding[idx[0], idx[1], idx[2]]

    ############################################################################

    def update_random_pool(self):

        """ Refills the random pool with new random numbers. """

        if self.rand_ctr >= self.max_rand:

            # self.write_log("Regenerating new random pool")
            for i in range(0, 3):
                self.random_pool[:, i] = self.random_generator.uniform(low=self.min_coord[i],
                                                                       high=self.max_coord[i],
                                                                       size=self.max_rand)

            self.rand_ctr = 0

    ############################################################################

    # density_function is either None, or a function of pos = [x,y,z] (in SI units)

    def define_density(self, neuron_type, density_function):

        """
        Defines density for neuron type.

        Args:
            neuron_type (str): Neuron type
            density_function (str): density_function is either None, or a function of pos = [x,y,z] (in SI units)

        """

        self.density_function[neuron_type] = density_function
        self.density_voxel_sum[neuron_type] = np.zeros(self.num_bins, dtype=float)
        self.density_total_sum[neuron_type] = 0
        self.density_voxel_n_sample[neuron_type] = np.zeros(self.num_bins, dtype=int)
        self.density_total_n_sample[neuron_type] = 0
        self.placed_voxel[neuron_type] = np.zeros(self.num_bins, dtype=int)
        self.placed_total[neuron_type] = 0

    def place_neurons(self, num_cells, neuron_type=None, d_min=None):

        if d_min is None:
            d_min = self.d_min

        d_min2 = d_min ** 2

        # If this is not fulfilled, then we need to update the range values below
        assert 2 * d_min < self.bin_width, \
            f"2*d_min (2 * {d_min}) must be smaller than bin_width ({self.bin_width})"

        if neuron_type in self.neuron_types:
            neuron_type_id = self.neuron_types[neuron_type]
        else:
            neuron_type_id = self.next_neuron_type
            self.next_neuron_type += 1

            self.neuron_types[neuron_type] = neuron_type_id

        start_ctr = self.neuron_ctr
        end_ctr = self.neuron_ctr + num_cells

        t_a = timeit.default_timer()

        if neuron_type in self.density_function and neuron_type != self.last_neuron_type_added:
            # Precalculate density function for potential putative locations
            xv = self.random_pool[self.rand_ctr:, 0]
            yv = self.random_pool[self.rand_ctr:, 1]
            zv = self.random_pool[self.rand_ctr:, 2]
            self.density_lookup[self.rand_ctr:] = self.density_function[neuron_type](x=xv, y=yv, z=zv)
            self.last_neuron_type_added = neuron_type

            if np.sum(self.density_lookup[self.rand_ctr:]) == 0:
                self.write_log(f"Density zero for {len(xv)} {neuron_type} neurons -- error with density?",
                               is_error=True)

        while self.neuron_ctr < end_ctr and self.reject_ctr < self.max_reject:

            putative_loc = self.random_pool[self.rand_ctr, :]
            df = self.density_lookup[self.rand_ctr]

            self.rand_ctr += 1

            if self.rand_ctr % 100000 == 0:
                self.write_log(f"Neurons: {self.neuron_ctr} Rejected: {self.reject_ctr} Padding: {self.padding_ctr}")

                if self.neuron_ctr == 0:
                    self.write_log("No neurons placed, check why!", is_error=True)
                    import pdb
                    pdb.set_trace()

            if self.rand_ctr >= self.max_rand:
                self.update_random_pool()

                if neuron_type in self.density_function:
                    xv = self.random_pool[:, 0]
                    yv = self.random_pool[:, 1]
                    zv = self.random_pool[:, 2]
                    self.density_lookup[:] = self.density_function[neuron_type](x=xv, y=yv, z=zv)

                # TODO: We should recalculate griddata density for these points

            inside_flag = self.check_inside(coords=putative_loc)

            if not inside_flag and not self.check_padding_zone(putative_loc):
                # We are outside, go to next neuron
                self.reject_ctr += 1
                continue

            # Check that we are not too close to existing points
            # Only check the neighbouring voxels, to speed things up
            voxel_idx = np.array(np.floor((putative_loc - self.min_coord) / self.bin_width), dtype=int)

            # Density check is fast, do that to get an early rejection if needed
            # TODO: When a function is defined it is relatively fast, but when griddata is used
            #       it is slooow...
            if inside_flag and neuron_type in self.density_function:
                # Update: df, density function value is now precomputed above
                # xp, yp, zp = putative_loc
                # df = self.density_function[neuron_type](x=xp, y=yp, z=zp)
                assert df >= 0, f"Error your density for {neuron_type} is incorrect, value={df} at {putative_loc}"
                vx, vy, vz = voxel_idx

                self.density_voxel_sum[neuron_type][vx, vy, vz] += df
                self.density_total_sum[neuron_type] += df
                self.density_voxel_n_sample[neuron_type][vx, vy, vz] += 1
                self.density_total_n_sample[neuron_type] += 1

                try:
                    # We are sampling the densities, to see how much of the density mass is in
                    # the current voxel compared to the total density mass in all voxels together
                    # That should tell us what fraction of all the neurons placed of that type
                    # should be in this particular voxel
                    n_expected = (self.density_voxel_sum[neuron_type][vx, vy, vz]
                                  / self.density_total_sum[neuron_type]
                                  * (self.placed_total[neuron_type] + 1))
                except:
                    import traceback
                    t_str = traceback.format_exc()
                    self.write_log(t_str)
                    import pdb
                    pdb.set_trace()

                # This assumes all of mesh voxels have same volume
                # n_expected = ((self.density_voxel_sum[neuron_type][vx, vy, vz]
                #               / self.density_voxel_n_sample[neuron_type][vx, vy, vz])
                #               / (self.density_total_sum[neuron_type] / self.density_total_n_sample[neuron_type])
                #               * (self.placed_total[neuron_type] + 1))

                if self.placed_voxel[neuron_type][vx, vy, vz] > np.ceil(n_expected):
                    # We have too many neurons in this part of the volume already, reject
                    self.reject_ctr += 1
                    continue

            voxel_idx_list = [voxel_idx]

            border_voxel = np.zeros((3,), dtype=int)

            for idx in range(0, 3):  # Looping over x,y,z coordinates
                if (putative_loc[idx] - self.min_coord[idx]) % self.bin_width < d_min:
                    border_voxel[idx] = -1
                    new_idx = np.copy(voxel_idx)
                    new_idx[idx] -= 1
                    voxel_idx_list.append(new_idx)

                elif (putative_loc[idx] - self.min_coord[idx]) % self.bin_width > self.bin_width - d_min:
                    border_voxel[idx] = 1
                    new_idx = np.copy(voxel_idx)
                    new_idx[idx] += 1
                    voxel_idx_list.append(new_idx)

            n_border = np.sum(np.abs(border_voxel))

            if n_border == 2:
                # Along one of the lines, need to check diagonal voxel also
                voxel_idx_list.append(voxel_idx + border_voxel)
            elif n_border == 3:
                # Close to corner, need to check 8 voxels in total (ouch!)
                voxel_idx_list.append(voxel_idx + [border_voxel[0], border_voxel[1], 0])
                voxel_idx_list.append(voxel_idx + [border_voxel[0], 0, border_voxel[2]])
                voxel_idx_list.append(voxel_idx + [0, border_voxel[1], border_voxel[2]])
                voxel_idx_list.append(voxel_idx + border_voxel)

            min_dist2 = 1e6
            for vox_idx in voxel_idx_list:
                if (vox_idx < 0).any() or (vox_idx > self.num_bins).any():
                    # Voxel outside bounds, ignore
                    continue

                if min_dist2 < d_min2:
                    # No need to calculate further, we are too close
                    break

                if self.voxel_next_neuron[vox_idx[0], vox_idx[1], vox_idx[2]] > 0:
                    tmp = self.voxel_neurons[vox_idx[0], vox_idx[1], vox_idx[2],
                          0:self.voxel_next_neuron[vox_idx[0], vox_idx[1], vox_idx[2]],
                          :] - putative_loc

                    min_dist2 = min(min_dist2, np.min(np.sum(np.square(tmp), axis=1)))

            if d_min2 < min_dist2:
                # Ok neuron is not too close to any neighbours

                if inside_flag:
                    # We are inside, add to inside points
                    self.neuron_coords[self.neuron_ctr, :] = putative_loc
                    self.neuron_type[self.neuron_ctr] = neuron_type_id
                    self.neuron_ctr += 1
                    # self.writeLog("Placed neuron " + str(self.neuronCtr))

                    # Update counts if we have a density function defined
                    if neuron_type in self.density_function:
                        self.placed_voxel[neuron_type][vx, vy, vz] += 1
                        self.placed_total[neuron_type] += 1
                else:
                    self.padding_ctr += 1

                # Also save the point in the specific voxel, this way we can ignore
                # lots of distance comparisons
                try:
                    self.voxel_neurons[voxel_idx[0], voxel_idx[1], voxel_idx[2],
                                       self.voxel_next_neuron[voxel_idx[0], voxel_idx[1], voxel_idx[2]],
                                       :] = putative_loc
                    self.voxel_next_neuron[voxel_idx[0], voxel_idx[1], voxel_idx[2]] += 1
                except:
                    self.write_log(f"If you see this error you probably need to increase "
                                   f"self.max_neurons_per_voxel={self.max_neurons_per_voxel}", is_error=True)
                    import traceback
                    tstr = traceback.format_exc()
                    print(tstr)
                    import pdb
                    pdb.set_trace()

            else:
                self.reject_ctr += 1

        t_b = timeit.default_timer()
        self.write_log(f"Placed {num_cells} in {t_b - t_a} s")

        if neuron_type in self.placed_voxel:
            if np.max(self.placed_voxel[neuron_type]) < 5 \
                    and neuron_type in self.density_function and self.density_function[neuron_type]:
                self.write_log(f"Warning, mesh_bin_width might be too small to setup accurate {neuron_type} density. "
                               f"Max {neuron_type} in any mesh voxel is {np.max(self.placed_voxel[neuron_type])}",
                               is_error=True)

        return self.neuron_coords[start_ctr:end_ctr, :]

        # Store point in global list, but also reference it in specific voxel

    ############################################################################

    # shape = "cube" or "sphere"
    # radius = radius of sphere, or half the length of side of the cube

    def get_subset(self, centre, radius=None, num_neurons=None, shape="cube",
                   return_idx_flag=False):

        """
        Returns subset of positions, either within a radius, or the closest num_neurons neurons.

        Args:
            centre (float,float,float): Centre of space
            radius (float): Radius if sphere, half-side if cube
            num_neurons (int): Number of neurons (if None, all within radius are returned)
            shape (str): "sphere" or "cube"
            return_idx_flag (bool): Return the indexes instead of coordinates (default False)

        """

        assert ((radius is None) ^ (num_neurons is None)), \
            "Specify one of radius or nNeurons."

        coords = self.neuron_coords[:self.neuron_ctr, :]
        nrn_type = self.neuron_type[:self.neuron_ctr, :]

        if shape == "cube":
            dist = np.amax(abs(coords - centre), axis=1)
        elif shape == "sphere":
            dist = np.sqrt(np.sum(np.square(coords - centre), axis=1))
        else:
            assert False, f"Unknown shape {shape} use cube or sphere"

        sorted_idx = np.argsort(dist)

        if radius is not None:
            self.write_log(f"Using radius {radius}")
            idx = sorted_idx[np.where(dist[sorted_idx] <= radius)]
        else:
            self.write_log(f"Selecting {num_neurons} closest neuron")
            idx = sorted_idx[:num_neurons]

        # Next we need to return them in the order they were originally sorted
        keep_mask = np.zeros((self.neuron_ctr,), dtype=bool)
        keep_mask[idx] = True

        if return_idx_flag:
            return np.where(keep_mask)
        else:
            return coords[keep_mask, :], nrn_type[keep_mask, :]

    ############################################################################

    def plot_struct(self, pdf_name=None):

        """ Plot structure.

        Args:
            pdf_name (str) : Save file name (default None)
        """

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        ax.scatter(self.mesh_vec[:, 0],
                   self.mesh_vec[:, 1],
                   self.mesh_vec[:, 2],
                   'black')

        plt.ion()
        plt.show()
        plt.pause(0.001)

        ax.view_init(150, 40)
        plt.draw()
        plt.axis("off")

        if pdf_name is not None:
            plt.savefig(pdf_name)

    ############################################################################

    def plot_neurons(self, plot_idx=None, pdf_name=None):

        """ Plot neurons.

        Args:
            plot_idx (list): Neuron ID to plot
            pdf_name (str): Name of file to save figure to (default None)
            """

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        if plot_idx is None:
            plot_idx = range(0, self.neuron_ctr)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        ax.scatter(self.neuron_coords[plot_idx, 0],
                   self.neuron_coords[plot_idx, 1],
                   self.neuron_coords[plot_idx, 2],
                   'black')

        plt.ion()
        plt.show()
        plt.pause(0.001)

        ax.view_init(150, 40)
        plt.draw()
        plt.axis("off")

        if pdf_name is not None:
            plt.savefig(pdf_name)

    ############################################################################

    def test_plot(self):

        """ Test plot"""

        # !!! NEXT ADD dMIN TO THIS

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        n_points = 300

        x_test = np.random.uniform(self.min_coord[0], self.max_coord[0], n_points)
        y_test = np.random.uniform(self.min_coord[1], self.max_coord[1], n_points)
        z_test = np.random.uniform(self.min_coord[2], self.max_coord[2], n_points)

        for i in range(0, n_points):

            self.write_log(f"Checking {i + 1}/{n_points}")

            if self.ray_casting(np.array([x_test[i], y_test[i], z_test[i]])):
                self.write_log("Inside!")
                color = 'red'
                ax.scatter(x_test[i], y_test[i], z_test[i], color=color)
            else:
                color = 'black'

            # ax.scatter(xTest[i],yTest[i],zTest[i],color=color)

        plt.show()
        plt.pause(0.001)

    ############################################################################

    def test_plot_cached(self):

        """ Test plot cached. """

        # !!! NEXT ADD dMIN TO THIS

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        n_points = 1000

        x_test = np.random.uniform(self.min_coord[0], self.max_coord[0], n_points)
        y_test = np.random.uniform(self.min_coord[1], self.max_coord[1], n_points)
        z_test = np.random.uniform(self.min_coord[2], self.max_coord[2], n_points)

        for i in range(0, n_points):

            coord = np.array([x_test[i], y_test[i], z_test[i]])
            ray_val = self.ray_casting(coord)
            cached_val = self.check_inside(coord)

            if ray_val == cached_val:
                self.write_log(f"Checking {i + 1}/{n_points}")

            elif ray_val:
                # Inside, but cached was wrong
                color = 'red'
                ax.scatter(x_test[i], y_test[i], z_test[i], color=color)
            else:
                # Outside, but cached wrong
                color = 'blue'
                ax.scatter(x_test[i], y_test[i], z_test[i], color=color)

        plt.show()
        plt.pause(0.001)

    ############################################################################

    def verify_d_min(self):

        """ Verify that d_min constraint is met. """

        self.write_log("Verifying that dMin constraint is met")

        min_dist = np.zeros((self.neuron_ctr,))

        if self.neuron_ctr < 100000:
            neuron_range = range(0, self.neuron_ctr)
        else:
            self.write_log("Too many to check all, picking random neurons to check")
            neuron_range = np.random.randint(0, self.neuron_ctr, size=(100000, 1))
            self.write_log(str(neuron_range))

        ctr = 0

        for i_neuron in range(0, self.neuron_ctr):

            ctr = ctr + 1
            if ctr % 10000 == 0:
                self.write_log(f"{ctr}/{len(neuron_range)}")

            d = np.sqrt(np.sum(np.square(self.neuron_coords[:self.neuron_ctr, :]
                                         - self.neuron_coords[i_neuron, :]),
                               axis=1))
            d[i_neuron] = 1e6  # Dont count self distance
            min_dist[i_neuron] = np.min(d)

        self.write_log(f"Closest neighbour, min = {np.min(min_dist)}")

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        if True:
            fig = plt.figure()
            plt.hist(min_dist)
            plt.title(f"Closest neighbour {np.min(min_dist)}")
            plt.ylabel("Count")
            plt.xlabel("Distance")
            plt.ion()
            plt.show()
            plt.pause(0.001)

        # Also plot where the neurons are relative to the voxel boundaries
        nr = [x for x in neuron_range]
        bad_idx = [nr[i] for i in np.where(min_dist < self.d_min)[0]]

        fig2 = plt.figure()
        ax = fig2.add_subplot(111, projection='3d')

        xc = (self.neuron_coords[bad_idx, :] - self.min_coord) % self.bin_width
        ax.scatter(xc[:, 0], xc[:, 1], xc[:, 2], 'black')
        plt.title(f"Bad neuron locations: {xc.shape[0]}")
        plt.axis('tight')
        plt.xlabel('X')
        plt.ylabel('Y')
        # plt.zlabel('Z')
        plt.ion()
        plt.show()
        plt.pause(0.001)

        try:
            assert self.d_min <= np.min(min_dist), \
                f"d_min criteria not fulfilled: {np.min(min_dist)} < d_min = {self.d_min}"
        except:
            self.write_log(f"d_min criteria not fulfilled: {np.min(min_dist)} < d_min = {self.d_min}")
            import pdb
            pdb.set_trace()

    ############################################################################

    def simple_test_case(self):

        self.write_log("This redefines the object, please restart after")

        self.filename = "cube.obj"
        self.cache_file = "cube.obj-cached.pickle"
        self.load_mesh(self.filename)
        self.pre_compute()
        self.setup_voxel_filter()

        self.point_out = np.array([0.5, 0.5, -10.5]) * 1e-6

        for i in range(0, 1000):
            test_point = np.random.uniform(0, 1e-6, 3) \
                         + np.array([0, 0, 0]) * 1e-6

            # testPoint = np.array([0.5,0.61,0.5])*1e-6

            # self.debugFlag = True
            inside_flag = self.ray_casting(test_point)

            if not inside_flag:
                self.write_log(f"Test point = {test_point}", is_error=True)
                self.write_log("wrong!", is_error=True)
                import pdb
                pdb.set_trace()

        self.write_log("All correct")
        self.write_log("Debug mode")
        import pdb
        pdb.set_trace()

    ############################################################################

    def write_log(self, text, flush=True, is_error=False, force_print=False):  # Change flush to False in future, debug

        """
        Writes to log file. Use setup_log first. Text is only written to screen if self.verbose=True,
        or is_error = True, or force_print = True.

        test (str) : Text to write
        flush (bool) : Should all writes be flushed to disk directly?
        is_error (bool) : Is this an error, always written.
        force_print (bool) : Force printing, even if self.verbose=False.
        """

        if self.logfile is not None:
            self.logfile.write(text + "\n")
            if flush:
                self.logfile.flush()

        if self.verbose or is_error or force_print:
            print(text)

    ############################################################################

    def inner_voxel_volume(self):

        """ Volume of all inner voxels. """

        return np.sum(self.voxel_mask_inner) * (self.bin_width ** 3)

    ############################################################################


if __name__ == "__main__":

    # sm = RegionMesh("cube.obj",useCache=False)
    # sm.simpleTestCase()

    if os.getenv('IPYTHON_PROFILE') is not None:
        from ipyparallel import Client

        rc = Client(profile=os.getenv('IPYTHON_PROFILE'),
                    # sshserver='127.0.0.1',
                    debug=False)
        print(f"Client IDs: {rc.ids}")

        # http://davidmasad.com/blog/simulation-with-ipyparallel/
        # http://people.duke.edu/~ccc14/sta-663-2016/19C_IPyParallel.html
        print(f"Client IDs: {rc.ids}")
        d_view = rc.direct_view(targets='all')  # rc[:] # Direct view into clients
    else:
        print("No IPYTHON_PROFILE enviroment variable set, running in serial")
        d_view = None
        rc = None

    meshFile = '../data/mesh/Striatum-d.obj'
    # meshFile = "mesh/cortex-mesh-200.obj"
    sm = RegionMesh(meshFile, d_view=d_view, raytrace_borders=False, verbose=True)

    # import cProfile
    # cProfile.run("neuronPos = sm.placeNeurons(1000)")

    # sm.plotStruct()

    nNeurons = 1730000
    neuronPos = sm.place_neurons(nNeurons)
    # sm.verify_d_min()
    sm.plot_neurons(pdf_name=os.path.join("figures", "striatum-fig-somas.png"))

    sm.plot_struct(pdf_name=os.path.join("figures", "striatum-fig-struct.png"))

    # sm.testPlot()
    # sm.testPlotCached()

    # tp = (sm.minCoord + sm.maxCoord)/2
    # sm.rayCasting(np.array(tp))

    if d_view and rc:
        rc.shutdown(hub=True)
